def abs(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_abs.html

    :param x: Argument.
    :type x: int32
    :rtype: int32
    """


def acos(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_acos.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def acosf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_acosf.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def acosh(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_acosh.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def acoshf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_acoshf.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def asin(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_asin.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def asinf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_asinf.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def asinh(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_asinh.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def asinhf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_asinhf.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def atan(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_atan.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def atan2(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_atan2.html

    :param x: Argument.
    :type x: float64
    :param y: Argument.
    :type y: float64
    :rtype: float64
    """


def atan2f(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_atan2f.html

    :param x: Argument.
    :type x: float32
    :param y: Argument.
    :type y: float32
    :rtype: float32
    """


def atanf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_atanf.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def atanh(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_atanh.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def atanhf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_atanhf.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def brev(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_brev.html

    :param x: Argument.
    :type x: int32
    :rtype: int32
    """


def brevll(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_brevll.html

    :param x: Argument.
    :type x: int64
    :rtype: int64
    """


def byte_perm(x, y, z):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_byte_perm.html

    :param x: Argument.
    :type x: int32
    :param y: Argument.
    :type y: int32
    :param z: Argument.
    :type z: int32
    :rtype: int32
    """


def cbrt(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_cbrt.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def cbrtf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_cbrtf.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def ceil(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ceil.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def ceilf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ceilf.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def clz(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_clz.html

    :param x: Argument.
    :type x: int32
    :rtype: int32
    """


def clzll(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_clzll.html

    :param x: Argument.
    :type x: int64
    :rtype: int32
    """


def copysign(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_copysign.html

    :param x: Argument.
    :type x: float64
    :param y: Argument.
    :type y: float64
    :rtype: float64
    """


def copysignf(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_copysignf.html

    :param x: Argument.
    :type x: float32
    :param y: Argument.
    :type y: float32
    :rtype: float32
    """


def cos(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_cos.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def cosf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_cosf.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def cosh(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_cosh.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def coshf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_coshf.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def cospi(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_cospi.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def cospif(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_cospif.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def dadd_rd(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_dadd_rd.html

    :param x: Argument.
    :type x: float64
    :param y: Argument.
    :type y: float64
    :rtype: float64
    """


def dadd_rn(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_dadd_rn.html

    :param x: Argument.
    :type x: float64
    :param y: Argument.
    :type y: float64
    :rtype: float64
    """


def dadd_ru(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_dadd_ru.html

    :param x: Argument.
    :type x: float64
    :param y: Argument.
    :type y: float64
    :rtype: float64
    """


def dadd_rz(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_dadd_rz.html

    :param x: Argument.
    :type x: float64
    :param y: Argument.
    :type y: float64
    :rtype: float64
    """


def ddiv_rd(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ddiv_rd.html

    :param x: Argument.
    :type x: float64
    :param y: Argument.
    :type y: float64
    :rtype: float64
    """


def ddiv_rn(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ddiv_rn.html

    :param x: Argument.
    :type x: float64
    :param y: Argument.
    :type y: float64
    :rtype: float64
    """


def ddiv_ru(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ddiv_ru.html

    :param x: Argument.
    :type x: float64
    :param y: Argument.
    :type y: float64
    :rtype: float64
    """


def ddiv_rz(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ddiv_rz.html

    :param x: Argument.
    :type x: float64
    :param y: Argument.
    :type y: float64
    :rtype: float64
    """


def dmul_rd(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_dmul_rd.html

    :param x: Argument.
    :type x: float64
    :param y: Argument.
    :type y: float64
    :rtype: float64
    """


def dmul_rn(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_dmul_rn.html

    :param x: Argument.
    :type x: float64
    :param y: Argument.
    :type y: float64
    :rtype: float64
    """


def dmul_ru(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_dmul_ru.html

    :param x: Argument.
    :type x: float64
    :param y: Argument.
    :type y: float64
    :rtype: float64
    """


def dmul_rz(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_dmul_rz.html

    :param x: Argument.
    :type x: float64
    :param y: Argument.
    :type y: float64
    :rtype: float64
    """


def double2float_rd(d):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double2float_rd.html

    :param d: Argument.
    :type d: float64
    :rtype: float32
    """


def double2float_rn(d):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double2float_rn.html

    :param d: Argument.
    :type d: float64
    :rtype: float32
    """


def double2float_ru(d):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double2float_ru.html

    :param d: Argument.
    :type d: float64
    :rtype: float32
    """


def double2float_rz(d):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double2float_rz.html

    :param d: Argument.
    :type d: float64
    :rtype: float32
    """


def double2hiint(d):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double2hiint.html

    :param d: Argument.
    :type d: float64
    :rtype: int32
    """


def double2int_rd(d):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double2int_rd.html

    :param d: Argument.
    :type d: float64
    :rtype: int32
    """


def double2int_rn(d):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double2int_rn.html

    :param d: Argument.
    :type d: float64
    :rtype: int32
    """


def double2int_ru(d):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double2int_ru.html

    :param d: Argument.
    :type d: float64
    :rtype: int32
    """


def double2int_rz(d):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double2int_rz.html

    :param d: Argument.
    :type d: float64
    :rtype: int32
    """


def double2ll_rd(f):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double2ll_rd.html

    :param f: Argument.
    :type f: float64
    :rtype: int64
    """


def double2ll_rn(f):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double2ll_rn.html

    :param f: Argument.
    :type f: float64
    :rtype: int64
    """


def double2ll_ru(f):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double2ll_ru.html

    :param f: Argument.
    :type f: float64
    :rtype: int64
    """


def double2ll_rz(f):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double2ll_rz.html

    :param f: Argument.
    :type f: float64
    :rtype: int64
    """


def double2loint(d):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double2loint.html

    :param d: Argument.
    :type d: float64
    :rtype: int32
    """


def double2uint_rd(d):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double2uint_rd.html

    :param d: Argument.
    :type d: float64
    :rtype: int32
    """


def double2uint_rn(d):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double2uint_rn.html

    :param d: Argument.
    :type d: float64
    :rtype: int32
    """


def double2uint_ru(d):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double2uint_ru.html

    :param d: Argument.
    :type d: float64
    :rtype: int32
    """


def double2uint_rz(d):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double2uint_rz.html

    :param d: Argument.
    :type d: float64
    :rtype: int32
    """


def double2ull_rd(f):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double2ull_rd.html

    :param f: Argument.
    :type f: float64
    :rtype: int64
    """


def double2ull_rn(f):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double2ull_rn.html

    :param f: Argument.
    :type f: float64
    :rtype: int64
    """


def double2ull_ru(f):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double2ull_ru.html

    :param f: Argument.
    :type f: float64
    :rtype: int64
    """


def double2ull_rz(f):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double2ull_rz.html

    :param f: Argument.
    :type f: float64
    :rtype: int64
    """


def double_as_longlong(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double_as_longlong.html

    :param x: Argument.
    :type x: float64
    :rtype: int64
    """


def drcp_rd(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_drcp_rd.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def drcp_rn(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_drcp_rn.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def drcp_ru(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_drcp_ru.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def drcp_rz(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_drcp_rz.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def dsqrt_rd(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_dsqrt_rd.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def dsqrt_rn(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_dsqrt_rn.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def dsqrt_ru(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_dsqrt_ru.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def dsqrt_rz(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_dsqrt_rz.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def erf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_erf.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def erfc(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_erfc.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def erfcf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_erfcf.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def erfcinv(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_erfcinv.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def erfcinvf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_erfcinvf.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def erfcx(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_erfcx.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def erfcxf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_erfcxf.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def erff(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_erff.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def erfinv(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_erfinv.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def erfinvf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_erfinvf.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def exp(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_exp.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def exp10(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_exp10.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def exp10f(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_exp10f.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def exp2(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_exp2.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def exp2f(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_exp2f.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def expf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_expf.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def expm1(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_expm1.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def expm1f(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_expm1f.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def fabs(f):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fabs.html

    :param f: Argument.
    :type f: float64
    :rtype: float64
    """


def fabsf(f):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fabsf.html

    :param f: Argument.
    :type f: float32
    :rtype: float32
    """


def fadd_rd(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fadd_rd.html

    :param x: Argument.
    :type x: float32
    :param y: Argument.
    :type y: float32
    :rtype: float32
    """


def fadd_rn(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fadd_rn.html

    :param x: Argument.
    :type x: float32
    :param y: Argument.
    :type y: float32
    :rtype: float32
    """


def fadd_ru(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fadd_ru.html

    :param x: Argument.
    :type x: float32
    :param y: Argument.
    :type y: float32
    :rtype: float32
    """


def fadd_rz(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fadd_rz.html

    :param x: Argument.
    :type x: float32
    :param y: Argument.
    :type y: float32
    :rtype: float32
    """


def fast_cosf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fast_cosf.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def fast_exp10f(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fast_exp10f.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def fast_expf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fast_expf.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def fast_fdividef(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fast_fdividef.html

    :param x: Argument.
    :type x: float32
    :param y: Argument.
    :type y: float32
    :rtype: float32
    """


def fast_log10f(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fast_log10f.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def fast_log2f(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fast_log2f.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def fast_logf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fast_logf.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def fast_powf(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fast_powf.html

    :param x: Argument.
    :type x: float32
    :param y: Argument.
    :type y: float32
    :rtype: float32
    """


def fast_sincosf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fast_sincosf.html

    :param x: Argument.
    :type x: float32
    :rtype: UniTuple(float32 x 2)
    """


def fast_sinf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fast_sinf.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def fast_tanf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fast_tanf.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def fdim(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fdim.html

    :param x: Argument.
    :type x: float64
    :param y: Argument.
    :type y: float64
    :rtype: float64
    """


def fdimf(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fdimf.html

    :param x: Argument.
    :type x: float32
    :param y: Argument.
    :type y: float32
    :rtype: float32
    """


def fdiv_rd(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fdiv_rd.html

    :param x: Argument.
    :type x: float32
    :param y: Argument.
    :type y: float32
    :rtype: float32
    """


def fdiv_rn(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fdiv_rn.html

    :param x: Argument.
    :type x: float32
    :param y: Argument.
    :type y: float32
    :rtype: float32
    """


def fdiv_ru(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fdiv_ru.html

    :param x: Argument.
    :type x: float32
    :param y: Argument.
    :type y: float32
    :rtype: float32
    """


def fdiv_rz(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fdiv_rz.html

    :param x: Argument.
    :type x: float32
    :param y: Argument.
    :type y: float32
    :rtype: float32
    """


def ffs(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ffs.html

    :param x: Argument.
    :type x: int32
    :rtype: int32
    """


def ffsll(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ffsll.html

    :param x: Argument.
    :type x: int64
    :rtype: int32
    """


def finitef(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_finitef.html

    :param x: Argument.
    :type x: float32
    :rtype: int32
    """


def float2half_rn(f):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_float2half_rn.html

    :param f: Argument.
    :type f: float32
    :rtype: int16
    """


def float2int_rd(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_float2int_rd.html

    :param in: Argument.
    :type in: float32
    :rtype: int32
    """


def float2int_rn(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_float2int_rn.html

    :param in: Argument.
    :type in: float32
    :rtype: int32
    """


def float2int_ru(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_float2int_ru.html

    :param in: Argument.
    :type in: float32
    :rtype: int32
    """


def float2int_rz(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_float2int_rz.html

    :param in: Argument.
    :type in: float32
    :rtype: int32
    """


def float2ll_rd(f):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_float2ll_rd.html

    :param f: Argument.
    :type f: float32
    :rtype: int64
    """


def float2ll_rn(f):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_float2ll_rn.html

    :param f: Argument.
    :type f: float32
    :rtype: int64
    """


def float2ll_ru(f):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_float2ll_ru.html

    :param f: Argument.
    :type f: float32
    :rtype: int64
    """


def float2ll_rz(f):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_float2ll_rz.html

    :param f: Argument.
    :type f: float32
    :rtype: int64
    """


def float2uint_rd(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_float2uint_rd.html

    :param in: Argument.
    :type in: float32
    :rtype: int32
    """


def float2uint_rn(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_float2uint_rn.html

    :param in: Argument.
    :type in: float32
    :rtype: int32
    """


def float2uint_ru(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_float2uint_ru.html

    :param in: Argument.
    :type in: float32
    :rtype: int32
    """


def float2uint_rz(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_float2uint_rz.html

    :param in: Argument.
    :type in: float32
    :rtype: int32
    """


def float2ull_rd(f):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_float2ull_rd.html

    :param f: Argument.
    :type f: float32
    :rtype: int64
    """


def float2ull_rn(f):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_float2ull_rn.html

    :param f: Argument.
    :type f: float32
    :rtype: int64
    """


def float2ull_ru(f):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_float2ull_ru.html

    :param f: Argument.
    :type f: float32
    :rtype: int64
    """


def float2ull_rz(f):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_float2ull_rz.html

    :param f: Argument.
    :type f: float32
    :rtype: int64
    """


def float_as_int(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_float_as_int.html

    :param x: Argument.
    :type x: float32
    :rtype: int32
    """


def floor(f):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_floor.html

    :param f: Argument.
    :type f: float64
    :rtype: float64
    """


def floorf(f):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_floorf.html

    :param f: Argument.
    :type f: float32
    :rtype: float32
    """


def fma(x, y, z):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fma.html

    :param x: Argument.
    :type x: float64
    :param y: Argument.
    :type y: float64
    :param z: Argument.
    :type z: float64
    :rtype: float64
    """


def fma_rd(x, y, z):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fma_rd.html

    :param x: Argument.
    :type x: float64
    :param y: Argument.
    :type y: float64
    :param z: Argument.
    :type z: float64
    :rtype: float64
    """


def fma_rn(x, y, z):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fma_rn.html

    :param x: Argument.
    :type x: float64
    :param y: Argument.
    :type y: float64
    :param z: Argument.
    :type z: float64
    :rtype: float64
    """


def fma_ru(x, y, z):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fma_ru.html

    :param x: Argument.
    :type x: float64
    :param y: Argument.
    :type y: float64
    :param z: Argument.
    :type z: float64
    :rtype: float64
    """


def fma_rz(x, y, z):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fma_rz.html

    :param x: Argument.
    :type x: float64
    :param y: Argument.
    :type y: float64
    :param z: Argument.
    :type z: float64
    :rtype: float64
    """


def fmaf(x, y, z):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fmaf.html

    :param x: Argument.
    :type x: float32
    :param y: Argument.
    :type y: float32
    :param z: Argument.
    :type z: float32
    :rtype: float32
    """


def fmaf_rd(x, y, z):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fmaf_rd.html

    :param x: Argument.
    :type x: float32
    :param y: Argument.
    :type y: float32
    :param z: Argument.
    :type z: float32
    :rtype: float32
    """


def fmaf_rn(x, y, z):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fmaf_rn.html

    :param x: Argument.
    :type x: float32
    :param y: Argument.
    :type y: float32
    :param z: Argument.
    :type z: float32
    :rtype: float32
    """


def fmaf_ru(x, y, z):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fmaf_ru.html

    :param x: Argument.
    :type x: float32
    :param y: Argument.
    :type y: float32
    :param z: Argument.
    :type z: float32
    :rtype: float32
    """


def fmaf_rz(x, y, z):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fmaf_rz.html

    :param x: Argument.
    :type x: float32
    :param y: Argument.
    :type y: float32
    :param z: Argument.
    :type z: float32
    :rtype: float32
    """


def fmax(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fmax.html

    :param x: Argument.
    :type x: float64
    :param y: Argument.
    :type y: float64
    :rtype: float64
    """


def fmaxf(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fmaxf.html

    :param x: Argument.
    :type x: float32
    :param y: Argument.
    :type y: float32
    :rtype: float32
    """


def fmin(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fmin.html

    :param x: Argument.
    :type x: float64
    :param y: Argument.
    :type y: float64
    :rtype: float64
    """


def fminf(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fminf.html

    :param x: Argument.
    :type x: float32
    :param y: Argument.
    :type y: float32
    :rtype: float32
    """


def fmod(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fmod.html

    :param x: Argument.
    :type x: float64
    :param y: Argument.
    :type y: float64
    :rtype: float64
    """


def fmodf(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fmodf.html

    :param x: Argument.
    :type x: float32
    :param y: Argument.
    :type y: float32
    :rtype: float32
    """


def fmul_rd(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fmul_rd.html

    :param x: Argument.
    :type x: float32
    :param y: Argument.
    :type y: float32
    :rtype: float32
    """


def fmul_rn(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fmul_rn.html

    :param x: Argument.
    :type x: float32
    :param y: Argument.
    :type y: float32
    :rtype: float32
    """


def fmul_ru(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fmul_ru.html

    :param x: Argument.
    :type x: float32
    :param y: Argument.
    :type y: float32
    :rtype: float32
    """


def fmul_rz(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fmul_rz.html

    :param x: Argument.
    :type x: float32
    :param y: Argument.
    :type y: float32
    :rtype: float32
    """


def frcp_rd(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_frcp_rd.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def frcp_rn(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_frcp_rn.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def frcp_ru(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_frcp_ru.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def frcp_rz(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_frcp_rz.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def frexp(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_frexp.html

    :param x: Argument.
    :type x: float64
    :rtype: Tuple(float64, int32)
    """


def frexpf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_frexpf.html

    :param x: Argument.
    :type x: float32
    :rtype: Tuple(float32, int32)
    """


def frsqrt_rn(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_frsqrt_rn.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def fsqrt_rd(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fsqrt_rd.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def fsqrt_rn(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fsqrt_rn.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def fsqrt_ru(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fsqrt_ru.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def fsqrt_rz(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fsqrt_rz.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def fsub_rd(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fsub_rd.html

    :param x: Argument.
    :type x: float32
    :param y: Argument.
    :type y: float32
    :rtype: float32
    """


def fsub_rn(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fsub_rn.html

    :param x: Argument.
    :type x: float32
    :param y: Argument.
    :type y: float32
    :rtype: float32
    """


def fsub_ru(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fsub_ru.html

    :param x: Argument.
    :type x: float32
    :param y: Argument.
    :type y: float32
    :rtype: float32
    """


def fsub_rz(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fsub_rz.html

    :param x: Argument.
    :type x: float32
    :param y: Argument.
    :type y: float32
    :rtype: float32
    """


def hadd(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_hadd.html

    :param x: Argument.
    :type x: int32
    :param y: Argument.
    :type y: int32
    :rtype: int32
    """


def half2float(h):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_half2float.html

    :param h: Argument.
    :type h: int16
    :rtype: float32
    """


def hiloint2double(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_hiloint2double.html

    :param x: Argument.
    :type x: int32
    :param y: Argument.
    :type y: int32
    :rtype: float64
    """


def hypot(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_hypot.html

    :param x: Argument.
    :type x: float64
    :param y: Argument.
    :type y: float64
    :rtype: float64
    """


def hypotf(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_hypotf.html

    :param x: Argument.
    :type x: float32
    :param y: Argument.
    :type y: float32
    :rtype: float32
    """


def ilogb(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ilogb.html

    :param x: Argument.
    :type x: float64
    :rtype: int32
    """


def ilogbf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ilogbf.html

    :param x: Argument.
    :type x: float32
    :rtype: int32
    """


def int2double_rn(i):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_int2double_rn.html

    :param i: Argument.
    :type i: int32
    :rtype: float64
    """


def int2float_rd(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_int2float_rd.html

    :param in: Argument.
    :type in: int32
    :rtype: float32
    """


def int2float_rn(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_int2float_rn.html

    :param in: Argument.
    :type in: int32
    :rtype: float32
    """


def int2float_ru(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_int2float_ru.html

    :param in: Argument.
    :type in: int32
    :rtype: float32
    """


def int2float_rz(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_int2float_rz.html

    :param in: Argument.
    :type in: int32
    :rtype: float32
    """


def int_as_float(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_int_as_float.html

    :param x: Argument.
    :type x: int32
    :rtype: float32
    """


def isfinited(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_isfinited.html

    :param x: Argument.
    :type x: float64
    :rtype: int32
    """


def isinfd(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_isinfd.html

    :param x: Argument.
    :type x: float64
    :rtype: int32
    """


def isinff(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_isinff.html

    :param x: Argument.
    :type x: float32
    :rtype: int32
    """


def isnand(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_isnand.html

    :param x: Argument.
    :type x: float64
    :rtype: int32
    """


def isnanf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_isnanf.html

    :param x: Argument.
    :type x: float32
    :rtype: int32
    """


def j0(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_j0.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def j0f(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_j0f.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def j1(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_j1.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def j1f(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_j1f.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def jn(n, x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_jn.html

    :param n: Argument.
    :type n: int32
    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def jnf(n, x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_jnf.html

    :param n: Argument.
    :type n: int32
    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def ldexp(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ldexp.html

    :param x: Argument.
    :type x: float64
    :param y: Argument.
    :type y: int32
    :rtype: float64
    """


def ldexpf(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ldexpf.html

    :param x: Argument.
    :type x: float32
    :param y: Argument.
    :type y: int32
    :rtype: float32
    """


def lgamma(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_lgamma.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def lgammaf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_lgammaf.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def ll2double_rd(l):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ll2double_rd.html

    :param l: Argument.
    :type l: int64
    :rtype: float64
    """


def ll2double_rn(l):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ll2double_rn.html

    :param l: Argument.
    :type l: int64
    :rtype: float64
    """


def ll2double_ru(l):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ll2double_ru.html

    :param l: Argument.
    :type l: int64
    :rtype: float64
    """


def ll2double_rz(l):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ll2double_rz.html

    :param l: Argument.
    :type l: int64
    :rtype: float64
    """


def ll2float_rd(l):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ll2float_rd.html

    :param l: Argument.
    :type l: int64
    :rtype: float32
    """


def ll2float_rn(l):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ll2float_rn.html

    :param l: Argument.
    :type l: int64
    :rtype: float32
    """


def ll2float_ru(l):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ll2float_ru.html

    :param l: Argument.
    :type l: int64
    :rtype: float32
    """


def ll2float_rz(l):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ll2float_rz.html

    :param l: Argument.
    :type l: int64
    :rtype: float32
    """


def llabs(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_llabs.html

    :param x: Argument.
    :type x: int64
    :rtype: int64
    """


def llmax(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_llmax.html

    :param x: Argument.
    :type x: int64
    :param y: Argument.
    :type y: int64
    :rtype: int64
    """


def llmin(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_llmin.html

    :param x: Argument.
    :type x: int64
    :param y: Argument.
    :type y: int64
    :rtype: int64
    """


def llrint(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_llrint.html

    :param x: Argument.
    :type x: float64
    :rtype: int64
    """


def llrintf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_llrintf.html

    :param x: Argument.
    :type x: float32
    :rtype: int64
    """


def llround(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_llround.html

    :param x: Argument.
    :type x: float64
    :rtype: int64
    """


def llroundf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_llroundf.html

    :param x: Argument.
    :type x: float32
    :rtype: int64
    """


def log(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_log.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def log10(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_log10.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def log10f(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_log10f.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def log1p(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_log1p.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def log1pf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_log1pf.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def log2(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_log2.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def log2f(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_log2f.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def logb(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_logb.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def logbf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_logbf.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def logf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_logf.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def longlong_as_double(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_longlong_as_double.html

    :param x: Argument.
    :type x: int64
    :rtype: float64
    """


def max(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_max.html

    :param x: Argument.
    :type x: int32
    :param y: Argument.
    :type y: int32
    :rtype: int32
    """


def min(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_min.html

    :param x: Argument.
    :type x: int32
    :param y: Argument.
    :type y: int32
    :rtype: int32
    """


def modf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_modf.html

    :param x: Argument.
    :type x: float64
    :rtype: UniTuple(float64 x 2)
    """


def modff(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_modff.html

    :param x: Argument.
    :type x: float32
    :rtype: UniTuple(float32 x 2)
    """


def mul24(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_mul24.html

    :param x: Argument.
    :type x: int32
    :param y: Argument.
    :type y: int32
    :rtype: int32
    """


def mul64hi(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_mul64hi.html

    :param x: Argument.
    :type x: int64
    :param y: Argument.
    :type y: int64
    :rtype: int64
    """


def mulhi(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_mulhi.html

    :param x: Argument.
    :type x: int32
    :param y: Argument.
    :type y: int32
    :rtype: int32
    """


def nearbyint(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_nearbyint.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def nearbyintf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_nearbyintf.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def nextafter(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_nextafter.html

    :param x: Argument.
    :type x: float64
    :param y: Argument.
    :type y: float64
    :rtype: float64
    """


def nextafterf(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_nextafterf.html

    :param x: Argument.
    :type x: float32
    :param y: Argument.
    :type y: float32
    :rtype: float32
    """


def normcdf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_normcdf.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def normcdff(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_normcdff.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def normcdfinv(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_normcdfinv.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def normcdfinvf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_normcdfinvf.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def popc(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_popc.html

    :param x: Argument.
    :type x: int32
    :rtype: int32
    """


def popcll(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_popcll.html

    :param x: Argument.
    :type x: int64
    :rtype: int32
    """


def pow(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_pow.html

    :param x: Argument.
    :type x: float64
    :param y: Argument.
    :type y: float64
    :rtype: float64
    """


def powf(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_powf.html

    :param x: Argument.
    :type x: float32
    :param y: Argument.
    :type y: float32
    :rtype: float32
    """


def powi(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_powi.html

    :param x: Argument.
    :type x: float64
    :param y: Argument.
    :type y: int32
    :rtype: float64
    """


def powif(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_powif.html

    :param x: Argument.
    :type x: float32
    :param y: Argument.
    :type y: int32
    :rtype: float32
    """


def rcbrt(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_rcbrt.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def rcbrtf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_rcbrtf.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def remainder(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_remainder.html

    :param x: Argument.
    :type x: float64
    :param y: Argument.
    :type y: float64
    :rtype: float64
    """


def remainderf(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_remainderf.html

    :param x: Argument.
    :type x: float32
    :param y: Argument.
    :type y: float32
    :rtype: float32
    """


def remquo(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_remquo.html

    :param x: Argument.
    :type x: float64
    :param y: Argument.
    :type y: float64
    :rtype: Tuple(float64, int32)
    """


def remquof(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_remquof.html

    :param x: Argument.
    :type x: float32
    :param y: Argument.
    :type y: float32
    :rtype: Tuple(float32, int32)
    """


def rhadd(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_rhadd.html

    :param x: Argument.
    :type x: int32
    :param y: Argument.
    :type y: int32
    :rtype: int32
    """


def rint(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_rint.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def rintf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_rintf.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def round(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_round.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def roundf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_roundf.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def rsqrt(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_rsqrt.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def rsqrtf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_rsqrtf.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def sad(x, y, z):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_sad.html

    :param x: Argument.
    :type x: int32
    :param y: Argument.
    :type y: int32
    :param z: Argument.
    :type z: int32
    :rtype: int32
    """


def saturatef(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_saturatef.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def scalbn(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_scalbn.html

    :param x: Argument.
    :type x: float64
    :param y: Argument.
    :type y: int32
    :rtype: float64
    """


def scalbnf(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_scalbnf.html

    :param x: Argument.
    :type x: float32
    :param y: Argument.
    :type y: int32
    :rtype: float32
    """


def signbitd(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_signbitd.html

    :param x: Argument.
    :type x: float64
    :rtype: int32
    """


def signbitf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_signbitf.html

    :param x: Argument.
    :type x: float32
    :rtype: int32
    """


def sin(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_sin.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def sincos(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_sincos.html

    :param x: Argument.
    :type x: float64
    :rtype: UniTuple(float64 x 2)
    """


def sincosf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_sincosf.html

    :param x: Argument.
    :type x: float32
    :rtype: UniTuple(float32 x 2)
    """


def sincospi(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_sincospi.html

    :param x: Argument.
    :type x: float64
    :rtype: UniTuple(float64 x 2)
    """


def sincospif(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_sincospif.html

    :param x: Argument.
    :type x: float32
    :rtype: UniTuple(float32 x 2)
    """


def sinf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_sinf.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def sinh(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_sinh.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def sinhf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_sinhf.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def sinpi(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_sinpi.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def sinpif(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_sinpif.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def sqrt(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_sqrt.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def sqrtf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_sqrtf.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def tan(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_tan.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def tanf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_tanf.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def tanh(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_tanh.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def tanhf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_tanhf.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def tgamma(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_tgamma.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def tgammaf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_tgammaf.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def trunc(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_trunc.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def truncf(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_truncf.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def uhadd(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_uhadd.html

    :param x: Argument.
    :type x: int32
    :param y: Argument.
    :type y: int32
    :rtype: int32
    """


def uint2double_rn(i):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_uint2double_rn.html

    :param i: Argument.
    :type i: int32
    :rtype: float64
    """


def uint2float_rd(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_uint2float_rd.html

    :param in: Argument.
    :type in: int32
    :rtype: float32
    """


def uint2float_rn(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_uint2float_rn.html

    :param in: Argument.
    :type in: int32
    :rtype: float32
    """


def uint2float_ru(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_uint2float_ru.html

    :param in: Argument.
    :type in: int32
    :rtype: float32
    """


def uint2float_rz(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_uint2float_rz.html

    :param in: Argument.
    :type in: int32
    :rtype: float32
    """


def ull2double_rd(l):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ull2double_rd.html

    :param l: Argument.
    :type l: int64
    :rtype: float64
    """


def ull2double_rn(l):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ull2double_rn.html

    :param l: Argument.
    :type l: int64
    :rtype: float64
    """


def ull2double_ru(l):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ull2double_ru.html

    :param l: Argument.
    :type l: int64
    :rtype: float64
    """


def ull2double_rz(l):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ull2double_rz.html

    :param l: Argument.
    :type l: int64
    :rtype: float64
    """


def ull2float_rd(l):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ull2float_rd.html

    :param l: Argument.
    :type l: int64
    :rtype: float32
    """


def ull2float_rn(l):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ull2float_rn.html

    :param l: Argument.
    :type l: int64
    :rtype: float32
    """


def ull2float_ru(l):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ull2float_ru.html

    :param l: Argument.
    :type l: int64
    :rtype: float32
    """


def ull2float_rz(l):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ull2float_rz.html

    :param l: Argument.
    :type l: int64
    :rtype: float32
    """


def ullmax(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ullmax.html

    :param x: Argument.
    :type x: int64
    :param y: Argument.
    :type y: int64
    :rtype: int64
    """


def ullmin(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ullmin.html

    :param x: Argument.
    :type x: int64
    :param y: Argument.
    :type y: int64
    :rtype: int64
    """


def umax(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_umax.html

    :param x: Argument.
    :type x: int32
    :param y: Argument.
    :type y: int32
    :rtype: int32
    """


def umin(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_umin.html

    :param x: Argument.
    :type x: int32
    :param y: Argument.
    :type y: int32
    :rtype: int32
    """


def umul24(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_umul24.html

    :param x: Argument.
    :type x: int32
    :param y: Argument.
    :type y: int32
    :rtype: int32
    """


def umul64hi(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_umul64hi.html

    :param x: Argument.
    :type x: int64
    :param y: Argument.
    :type y: int64
    :rtype: int64
    """


def umulhi(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_umulhi.html

    :param x: Argument.
    :type x: int32
    :param y: Argument.
    :type y: int32
    :rtype: int32
    """


def urhadd(x, y):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_urhadd.html

    :param x: Argument.
    :type x: int32
    :param y: Argument.
    :type y: int32
    :rtype: int32
    """


def usad(x, y, z):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_usad.html

    :param x: Argument.
    :type x: int32
    :param y: Argument.
    :type y: int32
    :param z: Argument.
    :type z: int32
    :rtype: int32
    """


def y0(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_y0.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def y0f(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_y0f.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def y1(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_y1.html

    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def y1f(x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_y1f.html

    :param x: Argument.
    :type x: float32
    :rtype: float32
    """


def yn(n, x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_yn.html

    :param n: Argument.
    :type n: int32
    :param x: Argument.
    :type x: float64
    :rtype: float64
    """


def ynf(n, x):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ynf.html

    :param n: Argument.
    :type n: int32
    :param x: Argument.
    :type x: float32
    :rtype: float32
    """



import multiprocessing
import os
from numba.core import config
from numba.cuda.cudadrv.runtime import runtime
from numba.cuda.testing import unittest, SerialMixin, skip_on_cudasim
from unittest.mock import patch


def set_visible_devices_and_check(q):
    try:
        from numba import cuda
        import os

        os.environ["CUDA_VISIBLE_DEVICES"] = "0"
        q.put(len(cuda.gpus.lst))
    except:  # noqa: E722
        # Sentinel value for error executing test code
        q.put(-1)


if config.ENABLE_CUDASIM:
    SUPPORTED_VERSIONS = ((-1, -1),)
else:
    SUPPORTED_VERSIONS = (
        (11, 0),
        (11, 1),
        (11, 2),
        (11, 3),
        (11, 4),
        (11, 5),
        (11, 6),
        (11, 7),
    )


class TestRuntime(unittest.TestCase):
    def test_is_supported_version_true(self):
        for v in SUPPORTED_VERSIONS:
            with patch.object(runtime, "get_version", return_value=v):
                self.assertTrue(runtime.is_supported_version())

    @skip_on_cudasim("The simulator always simulates a supported runtime")
    def test_is_supported_version_false(self):
        # Check with an old unsupported version and some potential future
        # versions
        for v in ((10, 2), (11, 8), (12, 0)):
            with patch.object(runtime, "get_version", return_value=v):
                self.assertFalse(runtime.is_supported_version())

    def test_supported_versions(self):
        self.assertEqual(SUPPORTED_VERSIONS, runtime.supported_versions)


class TestVisibleDevices(unittest.TestCase, SerialMixin):
    def test_visible_devices_set_after_import(self):
        # See Issue #6149. This test checks that we can set
        # CUDA_VISIBLE_DEVICES after importing Numba and have the value
        # reflected in the available list of GPUs. Prior to the fix for this
        # issue, Numba made a call to runtime.get_version() on import that
        # initialized the driver and froze the list of available devices before
        # CUDA_VISIBLE_DEVICES could be set by the user.

        # Avoid importing cuda at the top level so that
        # set_visible_devices_and_check gets to import it first in its process
        from numba import cuda

        if len(cuda.gpus.lst) in (0, 1):
            self.skipTest("This test requires multiple GPUs")

        if os.environ.get("CUDA_VISIBLE_DEVICES"):
            msg = "Cannot test when CUDA_VISIBLE_DEVICES already set"
            self.skipTest(msg)

        ctx = multiprocessing.get_context("spawn")
        q = ctx.Queue()
        p = ctx.Process(target=set_visible_devices_and_check, args=(q,))
        p.start()
        try:
            visible_gpu_count = q.get()
        finally:
            p.join()

        # Make an obvious distinction between an error running the test code
        # and an incorrect number of GPUs in the list
        msg = "Error running set_visible_devices_and_check"
        self.assertNotEqual(visible_gpu_count, -1, msg=msg)

        # The actual check that we see only one GPU
        self.assertEqual(visible_gpu_count, 1)


if __name__ == "__main__":
    unittest.main()



from ctypes import byref, c_char, c_char_p, c_int, c_size_t, c_void_p, POINTER
from enum import IntEnum
from numba.cuda.cudadrv.error import (
    NvrtcError,
    NvrtcBuiltinOperationFailure,
    NvrtcCompilationError,
    NvrtcSupportError,
)
from numba.cuda.cuda_paths import get_cuda_paths

import functools
import os
import threading
import warnings

# Opaque handle for compilation unit
nvrtc_program = c_void_p

# Result code
nvrtc_result = c_int


class NvrtcResult(IntEnum):
    NVRTC_SUCCESS = 0
    NVRTC_ERROR_OUT_OF_MEMORY = 1
    NVRTC_ERROR_PROGRAM_CREATION_FAILURE = 2
    NVRTC_ERROR_INVALID_INPUT = 3
    NVRTC_ERROR_INVALID_PROGRAM = 4
    NVRTC_ERROR_INVALID_OPTION = 5
    NVRTC_ERROR_COMPILATION = 6
    NVRTC_ERROR_BUILTIN_OPERATION_FAILURE = 7
    NVRTC_ERROR_NO_NAME_EXPRESSIONS_AFTER_COMPILATION = 8
    NVRTC_ERROR_NO_LOWERED_NAMES_BEFORE_COMPILATION = 9
    NVRTC_ERROR_NAME_EXPRESSION_NOT_VALID = 10
    NVRTC_ERROR_INTERNAL_ERROR = 11


_nvrtc_lock = threading.Lock()


class NvrtcProgram:
    """
    A class for managing the lifetime of nvrtcProgram instances. Instances of
    the class own an nvrtcProgram; when an instance is deleted, the underlying
    nvrtcProgram is destroyed using the appropriate NVRTC API.
    """

    def __init__(self, nvrtc, handle):
        self._nvrtc = nvrtc
        self._handle = handle

    @property
    def handle(self):
        return self._handle

    def __del__(self):
        if self._handle:
            self._nvrtc.destroy_program(self)


class NVRTC:
    """
    Provides a Pythonic interface to the NVRTC APIs, abstracting away the C API
    calls.

    The sole instance of this class is a process-wide singleton, similar to the
    NVVM interface. Initialization is protected by a lock and uses the standard
    (for Numba) open_cudalib function to load the NVRTC library.
    """

    _CU11_2ONLY_PROTOTYPES = {
        # nvrtcResult nvrtcGetNumSupportedArchs(int *numArchs);
        "nvrtcGetNumSupportedArchs": (nvrtc_result, POINTER(c_int)),
        # nvrtcResult nvrtcGetSupportedArchs(int *supportedArchs);
        "nvrtcGetSupportedArchs": (nvrtc_result, POINTER(c_int)),
    }

    _CU12ONLY_PROTOTYPES = {
        # nvrtcResult nvrtcGetLTOIRSize(nvrtcProgram prog, size_t *ltoSizeRet);
        "nvrtcGetLTOIRSize": (nvrtc_result, nvrtc_program, POINTER(c_size_t)),
        # nvrtcResult nvrtcGetLTOIR(nvrtcProgram prog, char *lto);
        "nvrtcGetLTOIR": (nvrtc_result, nvrtc_program, c_char_p),
    }

    _PROTOTYPES = {
        # nvrtcResult nvrtcVersion(int *major, int *minor)
        "nvrtcVersion": (nvrtc_result, POINTER(c_int), POINTER(c_int)),
        # nvrtcResult nvrtcCreateProgram(nvrtcProgram *prog,
        #                                const char *src,
        #                                const char *name,
        #                                int numHeaders,
        #                                const char * const *headers,
        #                                const char * const *includeNames)
        "nvrtcCreateProgram": (
            nvrtc_result,
            nvrtc_program,
            c_char_p,
            c_char_p,
            c_int,
            POINTER(c_char_p),
            POINTER(c_char_p),
        ),
        # nvrtcResult nvrtcDestroyProgram(nvrtcProgram *prog);
        "nvrtcDestroyProgram": (nvrtc_result, POINTER(nvrtc_program)),
        # nvrtcResult nvrtcCompileProgram(nvrtcProgram prog,
        #                                 int numOptions,
        #                                 const char * const *options)
        "nvrtcCompileProgram": (
            nvrtc_result,
            nvrtc_program,
            c_int,
            POINTER(c_char_p),
        ),
        # nvrtcResult nvrtcGetPTXSize(nvrtcProgram prog, size_t *ptxSizeRet);
        "nvrtcGetPTXSize": (nvrtc_result, nvrtc_program, POINTER(c_size_t)),
        # nvrtcResult nvrtcGetPTX(nvrtcProgram prog, char *ptx);
        "nvrtcGetPTX": (nvrtc_result, nvrtc_program, c_char_p),
        # nvrtcResult nvrtcGetCUBINSize(nvrtcProgram prog,
        #                               size_t *cubinSizeRet);
        "nvrtcGetCUBINSize": (nvrtc_result, nvrtc_program, POINTER(c_size_t)),
        # nvrtcResult nvrtcGetCUBIN(nvrtcProgram prog, char *cubin);
        "nvrtcGetCUBIN": (nvrtc_result, nvrtc_program, c_char_p),
        # nvrtcResult nvrtcGetProgramLogSize(nvrtcProgram prog,
        #                                    size_t *logSizeRet);
        "nvrtcGetProgramLogSize": (
            nvrtc_result,
            nvrtc_program,
            POINTER(c_size_t),
        ),
        # nvrtcResult nvrtcGetProgramLog(nvrtcProgram prog, char *log);
        "nvrtcGetProgramLog": (nvrtc_result, nvrtc_program, c_char_p),
    }

    # Singleton reference
    __INSTANCE = None

    def __new__(cls):
        with _nvrtc_lock:
            if cls.__INSTANCE is None:
                from numba.cuda.cudadrv.libs import open_cudalib

                cls.__INSTANCE = inst = object.__new__(cls)
                try:
                    lib = open_cudalib("nvrtc")
                except OSError as e:
                    cls.__INSTANCE = None
                    raise NvrtcSupportError("NVRTC cannot be loaded") from e

                from numba.cuda.cudadrv.runtime import get_version

                if get_version() >= (11, 2):
                    inst._PROTOTYPES |= inst._CU11_2ONLY_PROTOTYPES
                if get_version() >= (12, 0):
                    inst._PROTOTYPES |= inst._CU12ONLY_PROTOTYPES

                # Find & populate functions
                for name, proto in inst._PROTOTYPES.items():
                    func = getattr(lib, name)
                    func.restype = proto[0]
                    func.argtypes = proto[1:]

                    @functools.wraps(func)
                    def checked_call(*args, func=func, name=name):
                        error = func(*args)
                        if error == NvrtcResult.NVRTC_ERROR_COMPILATION:
                            raise NvrtcCompilationError()
                        elif (
                            error
                            == NvrtcResult.NVRTC_ERROR_BUILTIN_OPERATION_FAILURE
                        ):
                            raise NvrtcBuiltinOperationFailure()
                        elif error != NvrtcResult.NVRTC_SUCCESS:
                            try:
                                error_name = NvrtcResult(error).name
                            except ValueError:
                                error_name = (
                                    "Unknown nvrtc_result "
                                    f"(error code: {error})"
                                )
                            msg = f"Failed to call {name}: {error_name}"
                            raise NvrtcError(msg)

                    setattr(inst, name, checked_call)

        return cls.__INSTANCE

    def get_supported_archs(self):
        """
        Get Supported Architectures by NVRTC as list of arch tuples.
        """
        ver = self.get_version()
        if ver < (11, 0):
            raise RuntimeError(
                "Unsupported CUDA version. CUDA 11.0 or higher is required."
            )
        elif ver == (11, 0):
            return [
                (3, 0),
                (3, 2),
                (3, 5),
                (3, 7),
                (5, 0),
                (5, 2),
                (5, 3),
                (6, 0),
                (6, 1),
                (6, 2),
                (7, 0),
                (7, 2),
                (7, 5),
            ]
        elif ver == (11, 1):
            return [
                (3, 5),
                (3, 7),
                (5, 0),
                (5, 2),
                (5, 3),
                (6, 0),
                (6, 1),
                (6, 2),
                (7, 0),
                (7, 2),
                (7, 5),
                (8, 0),
            ]
        else:
            num = c_int()
            self.nvrtcGetNumSupportedArchs(byref(num))
            archs = (c_int * num.value)()
            self.nvrtcGetSupportedArchs(archs)
            return [(archs[i] // 10, archs[i] % 10) for i in range(num.value)]

    def get_version(self):
        """
        Get the NVRTC version as a tuple (major, minor).
        """
        major = c_int()
        minor = c_int()
        self.nvrtcVersion(byref(major), byref(minor))
        return major.value, minor.value

    def create_program(self, src, name):
        """
        Create an NVRTC program with managed lifetime.
        """
        if isinstance(src, str):
            src = src.encode()
        if isinstance(name, str):
            name = name.encode()

        handle = nvrtc_program()

        # The final three arguments are for passing the contents of headers -
        # this is not supported, so there are 0 headers and the header names
        # and contents are null.
        self.nvrtcCreateProgram(byref(handle), src, name, 0, None, None)
        return NvrtcProgram(self, handle)

    def compile_program(self, program, options):
        """
        Compile an NVRTC program. Compilation may fail due to a user error in
        the source; this function returns ``True`` if there is a compilation
        error and ``False`` on success.
        """
        # We hold a list of encoded options to ensure they can't be collected
        # prior to the call to nvrtcCompileProgram
        encoded_options = [opt.encode() for opt in options]
        option_pointers = [c_char_p(opt) for opt in encoded_options]
        c_options_type = c_char_p * len(options)
        c_options = c_options_type(*option_pointers)
        try:
            self.nvrtcCompileProgram(program.handle, len(options), c_options)
            return False
        except (NvrtcCompilationError, NvrtcBuiltinOperationFailure):
            return True

    def destroy_program(self, program):
        """
        Destroy an NVRTC program.
        """
        self.nvrtcDestroyProgram(byref(program.handle))

    def get_compile_log(self, program):
        """
        Get the compile log as a Python string.
        """
        log_size = c_size_t()
        self.nvrtcGetProgramLogSize(program.handle, byref(log_size))

        log = (c_char * log_size.value)()
        self.nvrtcGetProgramLog(program.handle, log)

        return log.value.decode()

    def get_ptx(self, program):
        """
        Get the compiled PTX as a Python string.
        """
        ptx_size = c_size_t()
        self.nvrtcGetPTXSize(program.handle, byref(ptx_size))

        ptx = (c_char * ptx_size.value)()
        self.nvrtcGetPTX(program.handle, ptx)

        return ptx.value.decode()

    def get_lto(self, program):
        """
        Get the compiled LTOIR as a Python bytes object.
        """
        lto_size = c_size_t()
        self.nvrtcGetLTOIRSize(program.handle, byref(lto_size))

        lto = b" " * lto_size.value
        self.nvrtcGetLTOIR(program.handle, lto)

        return lto


def compile(src, name, cc, ltoir=False):
    """
    Compile a CUDA C/C++ source to PTX or LTOIR for a given compute capability.

    :param src: The source code to compile
    :type src: str
    :param name: The filename of the source (for information only)
    :type name: str
    :param cc: A tuple ``(major, minor)`` of the compute capability
    :type cc: tuple
    :param ltoir: Compile into LTOIR if True, otherwise into PTX
    :type ltoir: bool
    :return: The compiled PTX and compilation log
    :rtype: tuple
    """
    nvrtc = NVRTC()
    program = nvrtc.create_program(src, name)

    version = nvrtc.get_version()
    ver_str = lambda v: ".".join(v)
    if version < (11, 0):
        raise RuntimeError(
            "Unsupported CUDA version. CUDA 11.0 or higher is required."
        )
    else:
        supported_arch = nvrtc.get_supported_archs()
        try:
            found = max(filter(lambda v: v <= cc, [v for v in supported_arch]))
        except ValueError:
            raise RuntimeError(
                f"Device compute capability {ver_str(cc)} is less than the "
                f"minimum supported by NVRTC {ver_str(version)}. Supported "
                "compute capabilities are "
                f"{', '.join([ver_str(v) for v in supported_arch])}."
            )

        if found != cc:
            warnings.warn(
                f"Device compute capability {ver_str(cc)} is not supported by "
                f"NVRTC {ver_str(version)}. Using {ver_str(found)} instead."
            )

    # Compilation options:
    # - Compile for the current device's compute capability.
    # - The CUDA include path is added.
    # - Relocatable Device Code (rdc) is needed to prevent device functions
    #   being optimized away.
    major, minor = found
    arch = f"--gpu-architecture=compute_{major}{minor}"

    cuda_include = [
        f"-I{get_cuda_paths()['include_dir'].info}",
    ]

    nvrtc_version = nvrtc.get_version()
    nvrtc_ver_major = nvrtc_version[0]

    cudadrv_path = os.path.dirname(os.path.abspath(__file__))
    numba_cuda_path = os.path.dirname(cudadrv_path)

    if nvrtc_ver_major == 11:
        numba_include = f"-I{os.path.join(numba_cuda_path, 'include', '11')}"
    else:
        numba_include = f"-I{os.path.join(numba_cuda_path, 'include', '12')}"

    nrt_path = os.path.join(numba_cuda_path, "runtime")
    nrt_include = f"-I{nrt_path}"

    options = [arch, numba_include, *cuda_include, nrt_include, "-rdc", "true"]

    if ltoir:
        options.append("-dlto")

    if nvrtc_version < (12, 0):
        options += ["-std=c++17"]

    # Compile the program
    compile_error = nvrtc.compile_program(program, options)

    # Get log from compilation
    log = nvrtc.get_compile_log(program)

    # If the compile failed, provide the log in an exception
    if compile_error:
        msg = f"NVRTC Compilation failure whilst compiling {name}:\n\n{log}"
        raise NvrtcError(msg)

    # Otherwise, if there's any content in the log, present it as a warning
    if log:
        msg = f"NVRTC log messages whilst compiling {name}:\n\n{log}"
        warnings.warn(msg)

    if ltoir:
        ltoir = nvrtc.get_lto(program)
        return ltoir, log
    else:
        ptx = nvrtc.get_ptx(program)
        return ptx, log



import numpy as np
from numba import cuda, int32, float32
from numba.cuda.testing import skip_on_cudasim, unittest, CUDATestCase
from numba.core.config import ENABLE_CUDASIM


def useless_syncthreads(ary):
    i = cuda.grid(1)
    cuda.syncthreads()
    ary[i] = i


def useless_syncwarp(ary):
    i = cuda.grid(1)
    cuda.syncwarp()
    ary[i] = i


def useless_syncwarp_with_mask(ary):
    i = cuda.grid(1)
    cuda.syncwarp(0xFFFF)
    ary[i] = i


def coop_syncwarp(res):
    sm = cuda.shared.array(32, int32)
    i = cuda.grid(1)

    sm[i] = i
    cuda.syncwarp()

    if i < 16:
        sm[i] = sm[i] + sm[i + 16]
        cuda.syncwarp(0xFFFF)

    if i < 8:
        sm[i] = sm[i] + sm[i + 8]
        cuda.syncwarp(0xFF)

    if i < 4:
        sm[i] = sm[i] + sm[i + 4]
        cuda.syncwarp(0xF)

    if i < 2:
        sm[i] = sm[i] + sm[i + 2]
        cuda.syncwarp(0x3)

    if i == 0:
        res[0] = sm[0] + sm[1]


def simple_smem(ary):
    N = 100
    sm = cuda.shared.array(N, int32)
    i = cuda.grid(1)
    if i == 0:
        for j in range(N):
            sm[j] = j
    cuda.syncthreads()
    ary[i] = sm[i]


def coop_smem2d(ary):
    i, j = cuda.grid(2)
    sm = cuda.shared.array((10, 20), float32)
    sm[i, j] = (i + 1) / (j + 1)
    cuda.syncthreads()
    ary[i, j] = sm[i, j]


def dyn_shared_memory(ary):
    i = cuda.grid(1)
    sm = cuda.shared.array(0, float32)
    sm[i] = i * 2
    cuda.syncthreads()
    ary[i] = sm[i]


def use_threadfence(ary):
    ary[0] += 123
    cuda.threadfence()
    ary[0] += 321


def use_threadfence_block(ary):
    ary[0] += 123
    cuda.threadfence_block()
    ary[0] += 321


def use_threadfence_system(ary):
    ary[0] += 123
    cuda.threadfence_system()
    ary[0] += 321


def use_syncthreads_count(ary_in, ary_out):
    i = cuda.grid(1)
    ary_out[i] = cuda.syncthreads_count(ary_in[i])


def use_syncthreads_and(ary_in, ary_out):
    i = cuda.grid(1)
    ary_out[i] = cuda.syncthreads_and(ary_in[i])


def use_syncthreads_or(ary_in, ary_out):
    i = cuda.grid(1)
    ary_out[i] = cuda.syncthreads_or(ary_in[i])


def _safe_cc_check(cc):
    if ENABLE_CUDASIM:
        return True
    else:
        return cuda.get_current_device().compute_capability >= cc


class TestCudaSync(CUDATestCase):
    def _test_useless(self, kernel):
        compiled = cuda.jit("void(int32[::1])")(kernel)
        nelem = 10
        ary = np.empty(nelem, dtype=np.int32)
        exp = np.arange(nelem, dtype=np.int32)
        compiled[1, nelem](ary)
        np.testing.assert_equal(ary, exp)

    def test_useless_syncthreads(self):
        self._test_useless(useless_syncthreads)

    @skip_on_cudasim("syncwarp not implemented on cudasim")
    def test_useless_syncwarp(self):
        self._test_useless(useless_syncwarp)

    @skip_on_cudasim("syncwarp not implemented on cudasim")
    @unittest.skipUnless(
        _safe_cc_check((7, 0)), "Partial masks require CC 7.0 or greater"
    )
    def test_useless_syncwarp_with_mask(self):
        self._test_useless(useless_syncwarp_with_mask)

    @skip_on_cudasim("syncwarp not implemented on cudasim")
    @unittest.skipUnless(
        _safe_cc_check((7, 0)), "Partial masks require CC 7.0 or greater"
    )
    def test_coop_syncwarp(self):
        # coop_syncwarp computes the sum of all integers from 0 to 31 (496)
        # using a single warp
        expected = 496
        nthreads = 32
        nblocks = 1

        compiled = cuda.jit("void(int32[::1])")(coop_syncwarp)
        res = np.zeros(1, dtype=np.int32)
        compiled[nblocks, nthreads](res)
        np.testing.assert_equal(expected, res[0])

    def test_simple_smem(self):
        compiled = cuda.jit("void(int32[::1])")(simple_smem)
        nelem = 100
        ary = np.empty(nelem, dtype=np.int32)
        compiled[1, nelem](ary)
        self.assertTrue(np.all(ary == np.arange(nelem, dtype=np.int32)))

    def test_coop_smem2d(self):
        compiled = cuda.jit("void(float32[:,::1])")(coop_smem2d)
        shape = 10, 20
        ary = np.empty(shape, dtype=np.float32)
        compiled[1, shape](ary)
        exp = np.empty_like(ary)
        for i in range(ary.shape[0]):
            for j in range(ary.shape[1]):
                exp[i, j] = (i + 1) / (j + 1)
        self.assertTrue(np.allclose(ary, exp))

    def test_dyn_shared_memory(self):
        compiled = cuda.jit("void(float32[::1])")(dyn_shared_memory)
        shape = 50
        ary = np.empty(shape, dtype=np.float32)
        compiled[1, shape, 0, ary.size * 4](ary)
        self.assertTrue(np.all(ary == 2 * np.arange(ary.size, dtype=np.int32)))

    def test_threadfence_codegen(self):
        # Does not test runtime behavior, just the code generation.
        sig = (int32[:],)
        compiled = cuda.jit(sig)(use_threadfence)
        ary = np.zeros(10, dtype=np.int32)
        compiled[1, 1](ary)
        self.assertEqual(123 + 321, ary[0])
        if not ENABLE_CUDASIM:
            self.assertIn("membar.gl;", compiled.inspect_asm(sig))

    def test_threadfence_block_codegen(self):
        # Does not test runtime behavior, just the code generation.
        sig = (int32[:],)
        compiled = cuda.jit(sig)(use_threadfence_block)
        ary = np.zeros(10, dtype=np.int32)
        compiled[1, 1](ary)
        self.assertEqual(123 + 321, ary[0])
        if not ENABLE_CUDASIM:
            self.assertIn("membar.cta;", compiled.inspect_asm(sig))

    def test_threadfence_system_codegen(self):
        # Does not test runtime behavior, just the code generation.
        sig = (int32[:],)
        compiled = cuda.jit(sig)(use_threadfence_system)
        ary = np.zeros(10, dtype=np.int32)
        compiled[1, 1](ary)
        self.assertEqual(123 + 321, ary[0])
        if not ENABLE_CUDASIM:
            self.assertIn("membar.sys;", compiled.inspect_asm(sig))

    def _test_syncthreads_count(self, in_dtype):
        compiled = cuda.jit(use_syncthreads_count)
        ary_in = np.ones(72, dtype=in_dtype)
        ary_out = np.zeros(72, dtype=np.int32)
        ary_in[31] = 0
        ary_in[42] = 0
        compiled[1, 72](ary_in, ary_out)
        self.assertTrue(np.all(ary_out == 70))

    def test_syncthreads_count(self):
        self._test_syncthreads_count(np.int32)

    def test_syncthreads_count_upcast(self):
        self._test_syncthreads_count(np.int16)

    def test_syncthreads_count_downcast(self):
        self._test_syncthreads_count(np.int64)

    def _test_syncthreads_and(self, in_dtype):
        compiled = cuda.jit(use_syncthreads_and)
        nelem = 100
        ary_in = np.ones(nelem, dtype=in_dtype)
        ary_out = np.zeros(nelem, dtype=np.int32)
        compiled[1, nelem](ary_in, ary_out)
        self.assertTrue(np.all(ary_out == 1))
        ary_in[31] = 0
        compiled[1, nelem](ary_in, ary_out)
        self.assertTrue(np.all(ary_out == 0))

    def test_syncthreads_and(self):
        self._test_syncthreads_and(np.int32)

    def test_syncthreads_and_upcast(self):
        self._test_syncthreads_and(np.int16)

    def test_syncthreads_and_downcast(self):
        self._test_syncthreads_and(np.int64)

    def _test_syncthreads_or(self, in_dtype):
        compiled = cuda.jit(use_syncthreads_or)
        nelem = 100
        ary_in = np.zeros(nelem, dtype=in_dtype)
        ary_out = np.zeros(nelem, dtype=np.int32)
        compiled[1, nelem](ary_in, ary_out)
        self.assertTrue(np.all(ary_out == 0))
        ary_in[31] = 1
        compiled[1, nelem](ary_in, ary_out)
        self.assertTrue(np.all(ary_out == 1))

    def test_syncthreads_or(self):
        self._test_syncthreads_or(np.int32)

    def test_syncthreads_or_upcast(self):
        self._test_syncthreads_or(np.int16)

    def test_syncthreads_or_downcast(self):
        self._test_syncthreads_or(np.int64)


if __name__ == "__main__":
    unittest.main()



import ctypes
import numpy as np
import weakref

from numba import cuda
from numba.core import config
from numba.cuda.testing import unittest, CUDATestCase, skip_on_cudasim
from numba.tests.support import linux_only

if not config.ENABLE_CUDASIM:

    class DeviceOnlyEMMPlugin(cuda.HostOnlyCUDAMemoryManager):
        """
        Dummy EMM Plugin implementation for testing. It memorises which plugin
        API methods have been called so that the tests can check that Numba
        called into the plugin as expected.
        """

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

            # For tracking our dummy allocations
            self.allocations = {}
            self.count = 0

            # For tracking which methods have been called
            self.initialized = False
            self.memalloc_called = False
            self.reset_called = False
            self.get_memory_info_called = False
            self.get_ipc_handle_called = False

        def memalloc(self, size):
            # We maintain a list of allocations and keep track of them, so that
            # we can test that the finalizers of objects returned by memalloc
            # get called.

            # Numba should have initialized the memory manager when preparing
            # the context for use, prior to any memalloc call.
            if not self.initialized:
                raise RuntimeError("memalloc called before initialize")
            self.memalloc_called = True

            # Create an allocation and record it
            self.count += 1
            alloc_count = self.count
            self.allocations[alloc_count] = size

            # The finalizer deletes the record from our internal dict of
            # allocations.
            finalizer_allocs = self.allocations

            def finalizer():
                del finalizer_allocs[alloc_count]

            # We use an AutoFreePointer so that the finalizer will be run when
            # the reference count drops to zero.
            ctx = weakref.proxy(self.context)
            ptr = ctypes.c_void_p(alloc_count)
            return cuda.cudadrv.driver.AutoFreePointer(
                ctx, ptr, size, finalizer=finalizer
            )

        def initialize(self):
            # No special initialization needed.
            self.initialized = True

        def reset(self):
            # We remove all allocations on reset, just as a real EMM Plugin
            # would do. Note that our finalizers in memalloc don't check
            # whether the allocations are still alive, so running them after
            # reset will detect any allocations that are floating around at
            # exit time; however, the atexit finalizer for weakref will only
            # print a traceback, not terminate the interpreter abnormally.
            self.reset_called = True

        def get_memory_info(self):
            # Return some dummy memory information
            self.get_memory_info_called = True
            return cuda.MemoryInfo(free=32, total=64)

        def get_ipc_handle(self, memory):
            # The dummy IPC handle is only a string, so it is important that
            # the tests don't try to do too much with it (e.g. open / close
            # it).
            self.get_ipc_handle_called = True
            return "Dummy IPC handle for alloc %s" % memory.device_pointer.value

        @property
        def interface_version(self):
            # The expected version for an EMM Plugin.
            return 1

    class BadVersionEMMPlugin(DeviceOnlyEMMPlugin):
        """A plugin that claims to implement a different interface version"""

        @property
        def interface_version(self):
            return 2


@skip_on_cudasim("EMM Plugins not supported on CUDA simulator")
class TestDeviceOnlyEMMPlugin(CUDATestCase):
    """
    Tests that the API of an EMM Plugin that implements device allocations
    only is used correctly by Numba.
    """

    def setUp(self):
        super().setUp()
        # Always start afresh with a new context and memory manager
        cuda.close()
        cuda.set_memory_manager(DeviceOnlyEMMPlugin)

    def tearDown(self):
        super().tearDown()
        # Unset the memory manager for subsequent tests
        cuda.close()
        cuda.cudadrv.driver._memory_manager = None

    def test_memalloc(self):
        mgr = cuda.current_context().memory_manager

        # Allocate an array and check that memalloc was called with the correct
        # size.
        arr_1 = np.arange(10)
        d_arr_1 = cuda.device_array_like(arr_1)
        self.assertTrue(mgr.memalloc_called)
        self.assertEqual(mgr.count, 1)
        self.assertEqual(mgr.allocations[1], arr_1.nbytes)

        # Allocate again, with a different size, and check that it is also
        # correct.
        arr_2 = np.arange(5)
        d_arr_2 = cuda.device_array_like(arr_2)
        self.assertEqual(mgr.count, 2)
        self.assertEqual(mgr.allocations[2], arr_2.nbytes)

        # Remove the first array, and check that our finalizer was called for
        # the first array only.
        del d_arr_1
        self.assertNotIn(1, mgr.allocations)
        self.assertIn(2, mgr.allocations)

        # Remove the second array and check that its finalizer was also
        # called.
        del d_arr_2
        self.assertNotIn(2, mgr.allocations)

    def test_initialized_in_context(self):
        # If we have a CUDA context, it should already have initialized its
        # memory manager.
        self.assertTrue(cuda.current_context().memory_manager.initialized)

    def test_reset(self):
        ctx = cuda.current_context()
        ctx.reset()
        self.assertTrue(ctx.memory_manager.reset_called)

    def test_get_memory_info(self):
        ctx = cuda.current_context()
        meminfo = ctx.get_memory_info()
        self.assertTrue(ctx.memory_manager.get_memory_info_called)
        self.assertEqual(meminfo.free, 32)
        self.assertEqual(meminfo.total, 64)

    @linux_only
    def test_get_ipc_handle(self):
        # We don't attempt to close the IPC handle in this test because Numba
        # will be expecting a real IpcHandle object to have been returned from
        # get_ipc_handle, and it would cause problems to do so.
        arr = np.arange(2)
        d_arr = cuda.device_array_like(arr)
        ipch = d_arr.get_ipc_handle()
        ctx = cuda.current_context()
        self.assertTrue(ctx.memory_manager.get_ipc_handle_called)
        self.assertIn("Dummy IPC handle for alloc 1", ipch._ipc_handle)


@skip_on_cudasim("EMM Plugins not supported on CUDA simulator")
class TestBadEMMPluginVersion(CUDATestCase):
    """
    Ensure that Numba rejects EMM Plugins with incompatible version
    numbers.
    """

    def test_bad_plugin_version(self):
        with self.assertRaises(RuntimeError) as raises:
            cuda.set_memory_manager(BadVersionEMMPlugin)
        self.assertIn("version 1 required", str(raises.exception))


if __name__ == "__main__":
    unittest.main()



import numpy as np
from numba.cuda.testing import (
    unittest,
    CUDATestCase,
    skip_unless_cc_53,
    skip_on_cudasim,
)
from numba import cuda
from numba.core.types import f2, b1
from numba.cuda import compile_ptx
import operator
import itertools
from numba.np.numpy_support import from_dtype


def simple_fp16_div_scalar(ary, a, b):
    ary[0] = a / b


def simple_fp16add(ary, a, b):
    ary[0] = a + b


def simple_fp16_iadd(ary, a):
    ary[0] += a


def simple_fp16_isub(ary, a):
    ary[0] -= a


def simple_fp16_imul(ary, a):
    ary[0] *= a


def simple_fp16_idiv(ary, a):
    ary[0] /= a


def simple_fp16sub(ary, a, b):
    ary[0] = a - b


def simple_fp16mul(ary, a, b):
    ary[0] = a * b


def simple_fp16neg(ary, a):
    ary[0] = -a


def simple_fp16abs(ary, a):
    ary[0] = abs(a)


def simple_fp16_gt(ary, a, b):
    ary[0] = a > b


def simple_fp16_ge(ary, a, b):
    ary[0] = a >= b


def simple_fp16_lt(ary, a, b):
    ary[0] = a < b


def simple_fp16_le(ary, a, b):
    ary[0] = a <= b


def simple_fp16_eq(ary, a, b):
    ary[0] = a == b


def simple_fp16_ne(ary, a, b):
    ary[0] = a != b


@cuda.jit("b1(f2, f2)", device=True)
def hlt_func_1(x, y):
    return x < y


@cuda.jit("b1(f2, f2)", device=True)
def hlt_func_2(x, y):
    return x < y


def test_multiple_hcmp_1(r, a, b, c):
    # float16 predicates used in two separate functions
    r[0] = hlt_func_1(a, b) and hlt_func_2(b, c)


def test_multiple_hcmp_2(r, a, b, c):
    # The same float16 predicate used in the caller and callee
    r[0] = hlt_func_1(a, b) and b < c


def test_multiple_hcmp_3(r, a, b, c):
    # Different float16 predicates used in the caller and callee
    r[0] = hlt_func_1(a, b) and c >= b


def test_multiple_hcmp_4(r, a, b, c):
    # The same float16 predicates used twice in a function
    r[0] = a < b and b < c


def test_multiple_hcmp_5(r, a, b, c):
    # Different float16 predicates used in a function
    r[0] = a < b and c >= b


class TestOperatorModule(CUDATestCase):
    def setUp(self):
        super().setUp()
        np.random.seed(0)

    """
    Test if operator module is supported by the CUDA target.
    """

    def operator_template(self, op):
        @cuda.jit
        def foo(a, b):
            i = 0
            a[i] = op(a[i], b[i])

        a = np.ones(1)
        b = np.ones(1)
        res = a.copy()
        foo[1, 1](res, b)

        np.testing.assert_equal(res, op(a, b))

    def test_add(self):
        self.operator_template(operator.add)

    def test_sub(self):
        self.operator_template(operator.sub)

    def test_mul(self):
        self.operator_template(operator.mul)

    def test_truediv(self):
        self.operator_template(operator.truediv)

    def test_floordiv(self):
        self.operator_template(operator.floordiv)

    @skip_unless_cc_53
    def test_fp16_binary(self):
        functions = (
            simple_fp16add,
            simple_fp16sub,
            simple_fp16mul,
            simple_fp16_div_scalar,
        )
        ops = (operator.add, operator.sub, operator.mul, operator.truediv)

        for fn, op in zip(functions, ops):
            with self.subTest(op=op):
                kernel = cuda.jit("void(f2[:], f2, f2)")(fn)

                got = np.zeros(1, dtype=np.float16)
                arg1 = np.random.random(1).astype(np.float16)
                arg2 = np.random.random(1).astype(np.float16)

                kernel[1, 1](got, arg1[0], arg2[0])
                expected = op(arg1, arg2)
                np.testing.assert_allclose(got, expected)

    @skip_on_cudasim("Compilation unsupported in the simulator")
    def test_fp16_binary_ptx(self):
        functions = (simple_fp16add, simple_fp16sub, simple_fp16mul)
        instrs = ("add.f16", "sub.f16", "mul.f16")
        args = (f2[:], f2, f2)
        for fn, instr in zip(functions, instrs):
            with self.subTest(instr=instr):
                ptx, _ = compile_ptx(fn, args, cc=(5, 3))
                self.assertIn(instr, ptx)

    @skip_unless_cc_53
    def test_mixed_fp16_binary_arithmetic(self):
        functions = (
            simple_fp16add,
            simple_fp16sub,
            simple_fp16mul,
            simple_fp16_div_scalar,
        )
        ops = (operator.add, operator.sub, operator.mul, operator.truediv)
        types = (np.int8, np.int16, np.int32, np.int64, np.float32, np.float64)
        for (fn, op), ty in itertools.product(zip(functions, ops), types):
            with self.subTest(op=op, ty=ty):
                kernel = cuda.jit(fn)

                arg1 = np.random.random(1).astype(np.float16)
                arg2 = (np.random.random(1) * 100).astype(ty)
                res_ty = np.result_type(np.float16, ty)

                got = np.zeros(1, dtype=res_ty)
                kernel[1, 1](got, arg1[0], arg2[0])
                expected = op(arg1, arg2)
                np.testing.assert_allclose(got, expected)

    @skip_on_cudasim("Compilation unsupported in the simulator")
    def test_fp16_inplace_binary_ptx(self):
        functions = (simple_fp16_iadd, simple_fp16_isub, simple_fp16_imul)
        instrs = ("add.f16", "sub.f16", "mul.f16")
        args = (f2[:], f2)

        for fn, instr in zip(functions, instrs):
            with self.subTest(instr=instr):
                ptx, _ = compile_ptx(fn, args, cc=(5, 3))
                self.assertIn(instr, ptx)

    @skip_unless_cc_53
    def test_fp16_inplace_binary(self):
        functions = (
            simple_fp16_iadd,
            simple_fp16_isub,
            simple_fp16_imul,
            simple_fp16_idiv,
        )
        ops = (operator.iadd, operator.isub, operator.imul, operator.itruediv)

        for fn, op in zip(functions, ops):
            with self.subTest(op=op):
                kernel = cuda.jit("void(f2[:], f2)")(fn)

                got = np.random.random(1).astype(np.float16)
                expected = got.copy()
                arg = np.random.random(1).astype(np.float16)[0]
                kernel[1, 1](got, arg)
                op(expected, arg)
                np.testing.assert_allclose(got, expected)

    @skip_unless_cc_53
    def test_fp16_unary(self):
        functions = (simple_fp16neg, simple_fp16abs)
        ops = (operator.neg, operator.abs)

        for fn, op in zip(functions, ops):
            with self.subTest(op=op):
                kernel = cuda.jit("void(f2[:], f2)")(fn)

                got = np.zeros(1, dtype=np.float16)
                arg1 = np.random.random(1).astype(np.float16)

                kernel[1, 1](got, arg1[0])
                expected = op(arg1)
                np.testing.assert_allclose(got, expected)

    @skip_on_cudasim("Compilation unsupported in the simulator")
    def test_fp16_neg_ptx(self):
        args = (f2[:], f2)
        ptx, _ = compile_ptx(simple_fp16neg, args, cc=(5, 3))
        self.assertIn("neg.f16", ptx)

    @skip_on_cudasim("Compilation unsupported in the simulator")
    def test_fp16_abs_ptx(self):
        args = (f2[:], f2)
        ptx, _ = compile_ptx(simple_fp16abs, args, cc=(5, 3))

        self.assertIn("abs.f16", ptx)

    @skip_unless_cc_53
    def test_fp16_comparison(self):
        functions = (
            simple_fp16_gt,
            simple_fp16_ge,
            simple_fp16_lt,
            simple_fp16_le,
            simple_fp16_eq,
            simple_fp16_ne,
        )
        ops = (
            operator.gt,
            operator.ge,
            operator.lt,
            operator.le,
            operator.eq,
            operator.ne,
        )

        for fn, op in zip(functions, ops):
            with self.subTest(op=op):
                kernel = cuda.jit("void(b1[:], f2, f2)")(fn)

                got = np.zeros(1, dtype=np.bool_)
                arg1 = np.random.random(1).astype(np.float16)
                arg2 = np.random.random(1).astype(np.float16)

                kernel[1, 1](got, arg1[0], arg2[0])
                expected = op(arg1, arg2)
                self.assertEqual(got[0], expected)

    @skip_unless_cc_53
    def test_mixed_fp16_comparison(self):
        functions = (
            simple_fp16_gt,
            simple_fp16_ge,
            simple_fp16_lt,
            simple_fp16_le,
            simple_fp16_eq,
            simple_fp16_ne,
        )
        ops = (
            operator.gt,
            operator.ge,
            operator.lt,
            operator.le,
            operator.eq,
            operator.ne,
        )
        types = (np.int8, np.int16, np.int32, np.int64, np.float32, np.float64)

        for (fn, op), ty in itertools.product(zip(functions, ops), types):
            with self.subTest(op=op, ty=ty):
                kernel = cuda.jit(fn)

                got = np.zeros(1, dtype=np.bool_)
                arg1 = np.random.random(1).astype(np.float16)
                arg2 = (np.random.random(1) * 100).astype(ty)

                kernel[1, 1](got, arg1[0], arg2[0])
                expected = op(arg1, arg2)
                self.assertEqual(got[0], expected)

    @skip_unless_cc_53
    def test_multiple_float16_comparisons(self):
        functions = (
            test_multiple_hcmp_1,
            test_multiple_hcmp_2,
            test_multiple_hcmp_3,
            test_multiple_hcmp_4,
            test_multiple_hcmp_5,
        )
        for fn in functions:
            with self.subTest(fn=fn):
                compiled = cuda.jit("void(b1[:], f2, f2, f2)")(fn)
                ary = np.zeros(1, dtype=np.bool_)
                arg1 = np.float16(2.0)
                arg2 = np.float16(3.0)
                arg3 = np.float16(4.0)
                compiled[1, 1](ary, arg1, arg2, arg3)
                self.assertTrue(ary[0])

    @skip_unless_cc_53
    def test_multiple_float16_comparisons_false(self):
        functions = (
            test_multiple_hcmp_1,
            test_multiple_hcmp_2,
            test_multiple_hcmp_3,
            test_multiple_hcmp_4,
            test_multiple_hcmp_5,
        )
        for fn in functions:
            with self.subTest(fn=fn):
                compiled = cuda.jit("void(b1[:], f2, f2, f2)")(fn)
                ary = np.zeros(1, dtype=np.bool_)
                arg1 = np.float16(2.0)
                arg2 = np.float16(3.0)
                arg3 = np.float16(1.0)
                compiled[1, 1](ary, arg1, arg2, arg3)
                self.assertFalse(ary[0])

    @skip_on_cudasim("Compilation unsupported in the simulator")
    def test_fp16_comparison_ptx(self):
        functions = (
            simple_fp16_gt,
            simple_fp16_ge,
            simple_fp16_lt,
            simple_fp16_le,
            simple_fp16_eq,
            simple_fp16_ne,
        )
        ops = (
            operator.gt,
            operator.ge,
            operator.lt,
            operator.le,
            operator.eq,
            operator.ne,
        )
        opstring = (
            "setp.gt.f16",
            "setp.ge.f16",
            "setp.lt.f16",
            "setp.le.f16",
            "setp.eq.f16",
            "setp.ne.f16",
        )
        args = (b1[:], f2, f2)

        for fn, op, s in zip(functions, ops, opstring):
            with self.subTest(op=op):
                ptx, _ = compile_ptx(fn, args, cc=(5, 3))
                self.assertIn(s, ptx)

    @skip_on_cudasim("Compilation unsupported in the simulator")
    def test_fp16_int8_comparison_ptx(self):
        # Test that int8 can be safely converted to fp16
        # in a comparison
        functions = (
            simple_fp16_gt,
            simple_fp16_ge,
            simple_fp16_lt,
            simple_fp16_le,
            simple_fp16_eq,
            simple_fp16_ne,
        )
        ops = (
            operator.gt,
            operator.ge,
            operator.lt,
            operator.le,
            operator.eq,
            operator.ne,
        )

        opstring = {
            operator.gt: "setp.gt.f16",
            operator.ge: "setp.ge.f16",
            operator.lt: "setp.lt.f16",
            operator.le: "setp.le.f16",
            operator.eq: "setp.eq.f16",
            operator.ne: "setp.ne.f16",
        }
        for fn, op in zip(functions, ops):
            with self.subTest(op=op):
                args = (b1[:], f2, from_dtype(np.int8))
                ptx, _ = compile_ptx(fn, args, cc=(5, 3))
                self.assertIn(opstring[op], ptx)

    @skip_on_cudasim("Compilation unsupported in the simulator")
    def test_mixed_fp16_comparison_promotion_ptx(self):
        functions = (
            simple_fp16_gt,
            simple_fp16_ge,
            simple_fp16_lt,
            simple_fp16_le,
            simple_fp16_eq,
            simple_fp16_ne,
        )
        ops = (
            operator.gt,
            operator.ge,
            operator.lt,
            operator.le,
            operator.eq,
            operator.ne,
        )

        types_promote = (np.int16, np.int32, np.int64, np.float32, np.float64)
        opstring = {
            operator.gt: "setp.gt.",
            operator.ge: "setp.ge.",
            operator.lt: "setp.lt.",
            operator.le: "setp.le.",
            operator.eq: "setp.eq.",
            operator.ne: "setp.neu.",
        }
        opsuffix = {
            np.dtype("int32"): "f64",
            np.dtype("int64"): "f64",
            np.dtype("float32"): "f32",
            np.dtype("float64"): "f64",
        }

        for (fn, op), ty in itertools.product(
            zip(functions, ops), types_promote
        ):
            with self.subTest(op=op, ty=ty):
                arg2_ty = np.result_type(np.float16, ty)
                args = (b1[:], f2, from_dtype(arg2_ty))
                ptx, _ = compile_ptx(fn, args, cc=(5, 3))

                ops = opstring[op] + opsuffix[arg2_ty]
                self.assertIn(ops, ptx)


if __name__ == "__main__":
    unittest.main()



from numba.cuda.testing import unittest
from numba.cuda.testing import skip_on_cudasim
from numba.cuda.testing import CUDATestCase
from numba.cuda.cudadrv.driver import PyNvJitLinker
from numba.cuda import get_current_device

from numba import cuda
from numba import config
from numba.tests.support import run_in_subprocess, override_config

try:
    import pynvjitlink  # noqa: F401

    PYNVJITLINK_INSTALLED = True
except ImportError:
    PYNVJITLINK_INSTALLED = False


import itertools
import os
import io
import contextlib
import warnings


TEST_BIN_DIR = os.getenv("NUMBA_CUDA_TEST_BIN_DIR")
if TEST_BIN_DIR:
    test_device_functions_a = os.path.join(
        TEST_BIN_DIR, "test_device_functions.a"
    )
    test_device_functions_cubin = os.path.join(
        TEST_BIN_DIR, "test_device_functions.cubin"
    )
    test_device_functions_cu = os.path.join(
        TEST_BIN_DIR, "test_device_functions.cu"
    )
    test_device_functions_fatbin = os.path.join(
        TEST_BIN_DIR, "test_device_functions.fatbin"
    )
    test_device_functions_fatbin_multi = os.path.join(
        TEST_BIN_DIR, "test_device_functions_multi.fatbin"
    )
    test_device_functions_o = os.path.join(
        TEST_BIN_DIR, "test_device_functions.o"
    )
    test_device_functions_ptx = os.path.join(
        TEST_BIN_DIR, "test_device_functions.ptx"
    )
    test_device_functions_ltoir = os.path.join(
        TEST_BIN_DIR, "test_device_functions.ltoir"
    )


@unittest.skipIf(
    not config.CUDA_ENABLE_PYNVJITLINK or not TEST_BIN_DIR,
    "pynvjitlink not enabled",
)
@skip_on_cudasim("Linking unsupported in the simulator")
class TestLinker(CUDATestCase):
    _NUMBA_NVIDIA_BINDING_0_ENV = {"NUMBA_CUDA_USE_NVIDIA_BINDING": "0"}

    def test_nvjitlink_create(self):
        patched_linker = PyNvJitLinker(cc=(7, 5))
        assert "-arch=sm_75" in patched_linker.options

    def test_nvjitlink_create_no_cc_error(self):
        # nvJitLink expects at least the architecture to be specified.
        with self.assertRaisesRegex(
            RuntimeError, "PyNvJitLinker requires CC to be specified"
        ):
            PyNvJitLinker()

    def test_nvjitlink_invalid_arch_error(self):
        from pynvjitlink.api import NvJitLinkError

        # CC 0.0 is not a valid compute capability
        with self.assertRaisesRegex(
            NvJitLinkError, "NVJITLINK_ERROR_UNRECOGNIZED_OPTION error"
        ):
            PyNvJitLinker(cc=(0, 0))

    def test_nvjitlink_invalid_cc_type_error(self):
        with self.assertRaisesRegex(
            TypeError, "`cc` must be a list or tuple of length 2"
        ):
            PyNvJitLinker(cc=0)

    def test_nvjitlink_ptx_compile_options(self):
        max_registers = (None, 32)
        lineinfo = (False, True)
        lto = (False, True)
        additional_flags = (None, ("-g",), ("-g", "-time"))
        for (
            max_registers_i,
            line_info_i,
            lto_i,
            additional_flags_i,
        ) in itertools.product(max_registers, lineinfo, lto, additional_flags):
            with self.subTest(
                max_registers=max_registers_i,
                lineinfo=line_info_i,
                lto=lto_i,
                additional_flags=additional_flags_i,
            ):
                patched_linker = PyNvJitLinker(
                    cc=(7, 5),
                    max_registers=max_registers_i,
                    lineinfo=line_info_i,
                    lto=lto_i,
                    additional_flags=additional_flags_i,
                )
                assert "-arch=sm_75" in patched_linker.options

                if max_registers_i:
                    assert (
                        f"-maxrregcount={max_registers_i}"
                        in patched_linker.options
                    )
                else:
                    assert "-maxrregcount" not in patched_linker.options

                if line_info_i:
                    assert "-lineinfo" in patched_linker.options
                else:
                    assert "-lineinfo" not in patched_linker.options

                if lto_i:
                    assert "-lto" in patched_linker.options
                else:
                    assert "-lto" not in patched_linker.options

                if additional_flags_i:
                    for flag in additional_flags_i:
                        assert flag in patched_linker.options

    def test_nvjitlink_add_file_guess_ext_linkable_code(self):
        files = (
            test_device_functions_a,
            test_device_functions_cubin,
            test_device_functions_cu,
            test_device_functions_fatbin,
            test_device_functions_o,
            test_device_functions_ptx,
        )
        for file in files:
            with self.subTest(file=file):
                patched_linker = PyNvJitLinker(
                    cc=get_current_device().compute_capability
                )
                patched_linker.add_file_guess_ext(file)

    def test_nvjitlink_test_add_file_guess_ext_invalid_input(self):
        with open(test_device_functions_cubin, "rb") as f:
            content = f.read()

        patched_linker = PyNvJitLinker(
            cc=get_current_device().compute_capability
        )
        with self.assertRaisesRegex(
            TypeError, "Expected path to file or a LinkableCode"
        ):
            # Feeding raw data as bytes to add_file_guess_ext should raise,
            # because there's no way to know what kind of file to treat it as
            patched_linker.add_file_guess_ext(content)

    def test_nvjitlink_jit_with_linkable_code(self):
        files = (
            test_device_functions_a,
            test_device_functions_cubin,
            test_device_functions_cu,
            test_device_functions_fatbin,
            test_device_functions_o,
            test_device_functions_ptx,
        )
        for lto in [True, False]:
            for file in files:
                with self.subTest(file=file):
                    sig = "uint32(uint32, uint32)"
                    add_from_numba = cuda.declare_device("add_from_numba", sig)

                    @cuda.jit(link=[file], lto=lto)
                    def kernel(result):
                        result[0] = add_from_numba(1, 2)

                    result = cuda.device_array(1)
                    kernel[1, 1](result)
                    assert result[0] == 3

    def test_nvjitlink_jit_with_linkable_code_lto_dump_assembly(self):
        files = [
            test_device_functions_cu,
            test_device_functions_ltoir,
            test_device_functions_fatbin_multi,
        ]

        config.DUMP_ASSEMBLY = True

        for file in files:
            with self.subTest(file=file):
                f = io.StringIO()
                with contextlib.redirect_stdout(f):
                    sig = "uint32(uint32, uint32)"
                    add_from_numba = cuda.declare_device("add_from_numba", sig)

                    @cuda.jit(link=[file], lto=True)
                    def kernel(result):
                        result[0] = add_from_numba(1, 2)

                    result = cuda.device_array(1)
                    kernel[1, 1](result)
                    assert result[0] == 3

                self.assertTrue("ASSEMBLY (AFTER LTO)" in f.getvalue())

        config.DUMP_ASSEMBLY = False

    def test_nvjitlink_jit_with_linkable_code_lto_dump_assembly_warn(self):
        files = [
            test_device_functions_a,
            test_device_functions_cubin,
            test_device_functions_fatbin,
            test_device_functions_o,
            test_device_functions_ptx,
        ]

        config.DUMP_ASSEMBLY = True

        for file in files:
            with self.subTest(file=file):
                with warnings.catch_warnings(record=True) as w:
                    with contextlib.redirect_stdout(None):  # suppress other PTX
                        sig = "uint32(uint32, uint32)"
                        add_from_numba = cuda.declare_device(
                            "add_from_numba", sig
                        )

                        @cuda.jit(link=[file], lto=True)
                        def kernel(result):
                            result[0] = add_from_numba(1, 2)

                        result = cuda.device_array(1)
                        kernel[1, 1](result)
                        assert result[0] == 3

                assert len(w) == 1
                self.assertIn(
                    "it is not optimizable at link time, and "
                    "`ignore_nonlto == True`",
                    str(w[0].message),
                )

        config.DUMP_ASSEMBLY = False

    def test_nvjitlink_jit_with_invalid_linkable_code(self):
        with open(test_device_functions_cubin, "rb") as f:
            content = f.read()
        with self.assertRaisesRegex(
            TypeError, "Expected path to file or a LinkableCode"
        ):

            @cuda.jit("void()", link=[content])
            def kernel():
                pass


@unittest.skipIf(
    not PYNVJITLINK_INSTALLED or not TEST_BIN_DIR,
    reason="pynvjitlink not enabled",
)
class TestLinkerUsage(CUDATestCase):
    """Test that whether pynvjitlink can be enabled by both environment variable
    and modification of config at runtime.
    """

    src = """if 1:
        import os
        from numba import cuda, config

        {config}

        TEST_BIN_DIR = os.getenv("NUMBA_CUDA_TEST_BIN_DIR")
        if TEST_BIN_DIR:
            test_device_functions_cubin = os.path.join(
                TEST_BIN_DIR, "test_device_functions.cubin"
            )

        sig = "uint32(uint32, uint32)"
        add_from_numba = cuda.declare_device("add_from_numba", sig)

        @cuda.jit(link=[test_device_functions_cubin], lto=True)
        def kernel(result):
            result[0] = add_from_numba(1, 2)

        result = cuda.device_array(1)
        kernel[1, 1](result)
        assert result[0] == 3
        """

    def test_linker_enabled_envvar(self):
        env = os.environ.copy()
        env["NUMBA_CUDA_ENABLE_PYNVJITLINK"] = "1"
        run_in_subprocess(self.src.format(config=""), env=env)

    def test_linker_disabled_envvar(self):
        env = os.environ.copy()
        env.pop("NUMBA_CUDA_ENABLE_PYNVJITLINK", None)
        with self.assertRaisesRegex(
            AssertionError, "LTO and additional flags require PyNvJitLinker"
        ):
            # Actual error raised is `ValueError`, but `run_in_subprocess`
            # reraises as AssertionError.
            run_in_subprocess(self.src.format(config=""), env=env)

    def test_linker_enabled_config(self):
        env = os.environ.copy()
        env.pop("NUMBA_CUDA_ENABLE_PYNVJITLINK", None)
        run_in_subprocess(
            self.src.format(config="config.CUDA_ENABLE_PYNVJITLINK = True"),
            env=env,
        )

    def test_linker_disabled_config(self):
        env = os.environ.copy()
        env.pop("NUMBA_CUDA_ENABLE_PYNVJITLINK", None)
        with override_config("CUDA_ENABLE_PYNVJITLINK", False):
            with self.assertRaisesRegex(
                AssertionError, "LTO and additional flags require PyNvJitLinker"
            ):
                run_in_subprocess(
                    self.src.format(
                        config="config.CUDA_ENABLE_PYNVJITLINK = False"
                    ),
                    env=env,
                )


if __name__ == "__main__":
    unittest.main()



import numpy as np

from numba.cuda import compile_ptx
from numba.core.types import f2, i1, i2, i4, i8, u1, u2, u4, u8
from numba import cuda
from numba.core import types
from numba.cuda.testing import CUDATestCase, skip_on_cudasim, skip_unless_cc_53
from numba.types import float16, float32
import itertools
import unittest


def native_cast(x):
    return float(x)


def to_int8(x):
    return np.int8(x)


def to_int16(x):
    return np.int16(x)


def to_int32(x):
    return np.int32(x)


def to_int64(x):
    return np.int64(x)


def to_uint8(x):
    return np.uint8(x)


def to_uint16(x):
    return np.uint16(x)


def to_uint32(x):
    return types.uint32(x)


def to_uint64(x):
    return types.uint64(x)


def to_float16(x):
    # When division and operators on float16 types are supported, this should
    # be changed to match the implementation in to_float32.
    return np.float16(x) * np.float16(0.5)


def to_float32(x):
    return np.float32(x) / np.float32(2)


def to_float64(x):
    return np.float64(x) / np.float64(2)


def to_complex64(x):
    return np.complex64(x)


def to_complex128(x):
    return np.complex128(x)


# Since multiplication of float16 is not supported via the operator * on
# float16s yet, and the host does not implement cuda.fp16.*, we need two
# versions of the following functions:
#
# - The device version uses cuda.fp16.hmul
# - The host version uses the * operator


def cuda_int_literal_to_float16(x):
    # Note that we need to use `2` and not `np.float16(2)` to ensure that this
    # types as a literal int and not a const float16.
    return cuda.fp16.hmul(np.float16(x), 2)


def reference_int_literal_to_float16(x):
    return np.float16(x) * np.float16(2)


def cuda_float_literal_to_float16(x):
    # Note that `2.5` types as a const float64 and not a literal float, but
    # this case is provided in case that changes in future.
    return cuda.fp16.hmul(np.float16(x), 2.5)


def reference_float_literal_to_float16(x):
    return np.float16(x) * np.float16(2.5)


class TestCasting(CUDATestCase):
    def _create_wrapped(self, pyfunc, intype, outtype):
        wrapped_func = cuda.jit(device=True)(pyfunc)

        @cuda.jit
        def cuda_wrapper_fn(arg, res):
            res[0] = wrapped_func(arg[0])

        def wrapper_fn(arg):
            argarray = np.zeros(1, dtype=intype)
            argarray[0] = arg
            resarray = np.zeros(1, dtype=outtype)
            cuda_wrapper_fn[1, 1](argarray, resarray)
            return resarray[0]

        return wrapper_fn

    @skip_unless_cc_53
    def test_float_to_int(self):
        pyfuncs = (to_int8, to_int16, to_int32, to_int64)
        totys = (np.int8, np.int16, np.int32, np.int64)
        fromtys = (np.float16, np.float32, np.float64)

        for pyfunc, toty in zip(pyfuncs, totys):
            for fromty in fromtys:
                with self.subTest(fromty=fromty, toty=toty):
                    cfunc = self._create_wrapped(pyfunc, fromty, toty)
                    self.assertEqual(cfunc(12.3), pyfunc(12.3))
                    self.assertEqual(cfunc(12.3), int(12.3))
                    self.assertEqual(cfunc(-12.3), pyfunc(-12.3))
                    self.assertEqual(cfunc(-12.3), int(-12.3))

    @skip_on_cudasim("Compilation unsupported in the simulator")
    def test_float16_to_int_ptx(self):
        pyfuncs = (to_int8, to_int16, to_int32, to_int64)
        sizes = (8, 16, 32, 64)

        for pyfunc, size in zip(pyfuncs, sizes):
            ptx, _ = compile_ptx(pyfunc, (f2,), device=True)
            self.assertIn(f"cvt.rni.s{size}.f16", ptx)

    @skip_unless_cc_53
    def test_float_to_uint(self):
        pyfuncs = (to_int8, to_int16, to_int32, to_int64)
        totys = (np.uint8, np.uint16, np.uint32, np.uint64)
        fromtys = (np.float16, np.float32, np.float64)

        for pyfunc, toty in zip(pyfuncs, totys):
            for fromty in fromtys:
                with self.subTest(fromty=fromty, toty=toty):
                    cfunc = self._create_wrapped(pyfunc, fromty, toty)
                    self.assertEqual(cfunc(12.3), pyfunc(12.3))
                    self.assertEqual(cfunc(12.3), int(12.3))

    @skip_on_cudasim("Compilation unsupported in the simulator")
    def test_float16_to_uint_ptx(self):
        pyfuncs = (to_uint8, to_uint16, to_uint32, to_uint64)
        sizes = (8, 16, 32, 64)

        for pyfunc, size in zip(pyfuncs, sizes):
            ptx, _ = compile_ptx(pyfunc, (f2,), device=True)
            self.assertIn(f"cvt.rni.u{size}.f16", ptx)

    @skip_unless_cc_53
    def test_int_to_float(self):
        pyfuncs = (to_float16, to_float32, to_float64)
        totys = (np.float16, np.float32, np.float64)

        for pyfunc, toty in zip(pyfuncs, totys):
            with self.subTest(toty=toty):
                cfunc = self._create_wrapped(pyfunc, np.int64, toty)
                self.assertEqual(cfunc(321), pyfunc(321))

    @skip_unless_cc_53
    def test_literal_to_float16(self):
        cudafuncs = (cuda_int_literal_to_float16, cuda_float_literal_to_float16)
        hostfuncs = (
            reference_int_literal_to_float16,
            reference_float_literal_to_float16,
        )

        for cudafunc, hostfunc in zip(cudafuncs, hostfuncs):
            with self.subTest(func=cudafunc):
                cfunc = self._create_wrapped(cudafunc, np.float16, np.float16)
                self.assertEqual(cfunc(321), hostfunc(321))

    @skip_on_cudasim("Compilation unsupported in the simulator")
    def test_int_to_float16_ptx(self):
        fromtys = (i1, i2, i4, i8)
        sizes = (8, 16, 32, 64)

        for ty, size in zip(fromtys, sizes):
            ptx, _ = compile_ptx(to_float16, (ty,), device=True)
            self.assertIn(f"cvt.rn.f16.s{size}", ptx)

    @skip_on_cudasim("Compilation unsupported in the simulator")
    def test_uint_to_float16_ptx(self):
        fromtys = (u1, u2, u4, u8)
        sizes = (8, 16, 32, 64)

        for ty, size in zip(fromtys, sizes):
            ptx, _ = compile_ptx(to_float16, (ty,), device=True)
            self.assertIn(f"cvt.rn.f16.u{size}", ptx)

    @skip_unless_cc_53
    def test_float_to_float(self):
        pyfuncs = (to_float16, to_float32, to_float64)
        tys = (np.float16, np.float32, np.float64)

        for (pyfunc, fromty), toty in itertools.product(zip(pyfuncs, tys), tys):
            with self.subTest(fromty=fromty, toty=toty):
                cfunc = self._create_wrapped(pyfunc, fromty, toty)
                # For this test we cannot use the pyfunc for comparison because
                # the CUDA target doesn't yet implement division (or operators)
                # for float16 values, so we test by comparing with the computed
                # expression instead.
                np.testing.assert_allclose(
                    cfunc(12.3), toty(12.3) / toty(2), rtol=0.0003
                )
                np.testing.assert_allclose(
                    cfunc(-12.3), toty(-12.3) / toty(2), rtol=0.0003
                )

    @skip_on_cudasim("Compilation unsupported in the simulator")
    def test_float16_to_float_ptx(self):
        pyfuncs = (to_float32, to_float64)
        postfixes = ("f32", "f64")

        for pyfunc, postfix in zip(pyfuncs, postfixes):
            ptx, _ = compile_ptx(pyfunc, (f2,), device=True)
            self.assertIn(f"cvt.{postfix}.f16", ptx)

    @skip_unless_cc_53
    def test_float_to_complex(self):
        pyfuncs = (to_complex64, to_complex128)
        totys = (np.complex64, np.complex128)
        fromtys = (np.float16, np.float32, np.float64)

        for pyfunc, toty in zip(pyfuncs, totys):
            for fromty in fromtys:
                with self.subTest(fromty=fromty, toty=toty):
                    cfunc = self._create_wrapped(pyfunc, fromty, toty)
                    # Here we need to explicitly cast the input to the pyfunc
                    # to match the casting that is automatically applied when
                    # passing the input to the cfunc as part of wrapping it in
                    # an array of type fromtype.
                    np.testing.assert_allclose(
                        cfunc(3.21), pyfunc(fromty(3.21))
                    )
                    np.testing.assert_allclose(
                        cfunc(-3.21), pyfunc(fromty(-3.21)) + 0j
                    )

    @skip_on_cudasim("Compilation unsupported in the simulator")
    def test_native_cast(self):
        float32_ptx, _ = cuda.compile_ptx(native_cast, (float32,), device=True)
        self.assertIn("st.f32", float32_ptx)

        float16_ptx, _ = cuda.compile_ptx(native_cast, (float16,), device=True)
        self.assertIn("st.u16", float16_ptx)


if __name__ == "__main__":
    unittest.main()



import threading

import numpy as np

from numba import cuda
from numba.cuda.testing import CUDATestCase, skip_unless_cudasim
import numba.cuda.simulator as simulator
import unittest


class TestCudaSimIssues(CUDATestCase):
    def test_record_access(self):
        backyard_type = [
            ("statue", np.float64),
            ("newspaper", np.float64, (6,)),
        ]

        goose_type = [
            ("garden", np.float64, (12,)),
            ("town", np.float64, (42,)),
            ("backyard", backyard_type),
        ]

        goose_np_type = np.dtype(goose_type, align=True)

        @cuda.jit
        def simple_kernel(f):
            f.garden[0] = 45.0
            f.backyard.newspaper[3] = 2.0
            f.backyard.newspaper[3] = f.backyard.newspaper[3] + 3.0

        item = np.recarray(1, dtype=goose_np_type)
        simple_kernel[1, 1](item[0])
        np.testing.assert_equal(item[0]["garden"][0], 45)
        np.testing.assert_equal(item[0]["backyard"]["newspaper"][3], 5)

    def test_recarray_setting(self):
        recordwith2darray = np.dtype(
            [("i", np.int32), ("j", np.float32, (3, 2))]
        )
        rec = np.recarray(2, dtype=recordwith2darray)
        rec[0]["i"] = 45

        @cuda.jit
        def simple_kernel(f):
            f[1] = f[0]

        simple_kernel[1, 1](rec)
        np.testing.assert_equal(rec[0]["i"], rec[1]["i"])

    def test_cuda_module_in_device_function(self):
        """
        Discovered in https://github.com/numba/numba/issues/1837.
        When the `cuda` module is referenced in a device function,
        it does not have the kernel API (e.g. cuda.threadIdx, cuda.shared)
        """
        from numba.cuda.tests.cudasim import support

        inner = support.cuda_module_in_device_function

        @cuda.jit
        def outer(out):
            tid = inner()
            if tid < out.size:
                out[tid] = tid

        arr = np.zeros(10, dtype=np.int32)
        outer[1, 11](arr)
        expected = np.arange(arr.size, dtype=np.int32)
        np.testing.assert_equal(expected, arr)

    @skip_unless_cudasim("Only works on CUDASIM")
    def test_deadlock_on_exception(self):
        def assert_no_blockthreads():
            blockthreads = []
            for t in threading.enumerate():
                if not isinstance(t, simulator.kernel.BlockThread):
                    continue

                # join blockthreads with a short timeout to allow aborted
                # threads to exit
                t.join(1)
                if t.is_alive():
                    self.fail("Blocked kernel thread: %s" % t)

            self.assertListEqual(blockthreads, [])

        @simulator.jit
        def assign_with_sync(x, y):
            i = cuda.grid(1)
            y[i] = x[i]

            cuda.syncthreads()
            cuda.syncthreads()

        x = np.arange(3)
        y = np.empty(3)
        assign_with_sync[1, 3](x, y)
        np.testing.assert_array_equal(x, y)
        assert_no_blockthreads()

        with self.assertRaises(IndexError):
            assign_with_sync[1, 6](x, y)
        assert_no_blockthreads()


if __name__ == "__main__":
    unittest.main()



import numpy as np
from numba.cuda.testing import unittest, CUDATestCase
from numba import cuda


def boolean_func(A, vertial):
    if vertial:
        A[0] = 123
    else:
        A[0] = 321


class TestCudaBoolean(CUDATestCase):
    def test_boolean(self):
        func = cuda.jit("void(float64[:], bool_)")(boolean_func)
        A = np.array([0], dtype="float64")
        func[1, 1](A, True)
        self.assertTrue(A[0] == 123)
        func[1, 1](A, False)
        self.assertTrue(A[0] == 321)


if __name__ == "__main__":
    unittest.main()



from numba.core.descriptors import TargetDescriptor
from numba.core.options import TargetOptions
from .target import CUDATargetContext, CUDATypingContext


class CUDATargetOptions(TargetOptions):
    pass


class CUDATarget(TargetDescriptor):
    def __init__(self, name):
        self.options = CUDATargetOptions
        # The typing and target contexts are initialized only when needed -
        # this prevents an attempt to load CUDA libraries at import time on
        # systems that might not have them present.
        self._typingctx = None
        self._targetctx = None
        super().__init__(name)

    @property
    def typing_context(self):
        if self._typingctx is None:
            self._typingctx = CUDATypingContext()
        return self._typingctx

    @property
    def target_context(self):
        if self._targetctx is None:
            self._targetctx = CUDATargetContext(self._typingctx)
        return self._targetctx


cuda_target = CUDATarget("cuda")



"""
Most of the driver API is unsupported in the simulator, but some stubs are
provided to allow tests to import correctly.
"""


def device_memset(dst, val, size, stream=0):
    dst.view("u1")[:size].fill(bytes([val])[0])


def host_to_device(dst, src, size, stream=0):
    dst.view("u1")[:size] = src.view("u1")[:size]


def device_to_host(dst, src, size, stream=0):
    host_to_device(dst, src, size)


def device_memory_size(obj):
    return obj.itemsize * obj.size


def device_to_device(dst, src, size, stream=0):
    host_to_device(dst, src, size)


class FakeDriver(object):
    def get_device_count(self):
        return 1


driver = FakeDriver()


class Linker:
    @classmethod
    def new(cls, max_registers=0, lineinfo=False, cc=None):
        return Linker()

    @property
    def lto(self):
        return False


class LinkerError(RuntimeError):
    pass


class NvrtcError(RuntimeError):
    pass


class CudaAPIError(RuntimeError):
    pass


def launch_kernel(*args, **kwargs):
    msg = "Launching kernels directly is not supported in the simulator"
    raise RuntimeError(msg)


USE_NV_BINDING = False



import os
import warnings
import traceback
import functools


def _readenv(name, ctor, default):
    value = os.environ.get(name)
    if value is None:
        return default() if callable(default) else default
    try:
        if ctor is bool:
            return value.lower() in {"1", "true"}
        return ctor(value)
    except Exception:
        warnings.warn(
            f"Environment variable '{name}' is defined but its associated "
            f"value '{value}' could not be parsed.\n"
            "The parse failed with exception:\n"
            f"{traceback.format_exc()}",
            RuntimeWarning,
        )
        return default


@functools.lru_cache(maxsize=None)
def cached_file_read(filepath, how="r"):
    with open(filepath, how) as f:
        return f.read()



from numba import cuda
from numba.cuda.testing import unittest, CUDATestCase

import numpy as np


class TestIterators(CUDATestCase):
    def test_enumerate(self):
        @cuda.jit
        def enumerator(x, error):
            count = 0

            for i, v in enumerate(x):
                if count != i:
                    error[0] = 1
                if v != x[i]:
                    error[0] = 2

                count += 1

            if count != len(x):
                error[0] = 3

        x = np.asarray((10, 9, 8, 7, 6))
        error = np.zeros(1, dtype=np.int32)

        enumerator[1, 1](x, error)
        self.assertEqual(error[0], 0)

    def _test_twoarg_function(self, f):
        x = np.asarray((10, 9, 8, 7, 6))
        y = np.asarray((1, 2, 3, 4, 5))
        error = np.zeros(1, dtype=np.int32)

        f[1, 1](x, y, error)
        self.assertEqual(error[0], 0)

    def test_zip(self):
        @cuda.jit
        def zipper(x, y, error):
            i = 0

            for xv, yv in zip(x, y):
                if xv != x[i]:
                    error[0] = 1
                if yv != y[i]:
                    error[0] = 2

                i += 1

            if i != len(x):
                error[0] = 3

        self._test_twoarg_function(zipper)

    def test_enumerate_zip(self):
        @cuda.jit
        def enumerator_zipper(x, y, error):
            count = 0

            for i, (xv, yv) in enumerate(zip(x, y)):
                if i != count:
                    error[0] = 1
                if xv != x[i]:
                    error[0] = 2
                if yv != y[i]:
                    error[0] = 3

                count += 1

            if count != len(x):
                error[0] = 4

        self._test_twoarg_function(enumerator_zipper)

    def test_zip_enumerate(self):
        @cuda.jit
        def zipper_enumerator(x, y, error):
            count = 0

            for (i, xv), yv in zip(enumerate(x), y):
                if i != count:
                    error[0] = 1
                if xv != x[i]:
                    error[0] = 2
                if yv != y[i]:
                    error[0] = 3

                count += 1

            if count != len(x):
                error[0] = 4

        self._test_twoarg_function(zipper_enumerator)


if __name__ == "__main__":
    unittest.main()
